# Copyright (c) 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.cli`.'''

import sys
import unittest
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from swift_scality_backend import cli
from swift_scality_backend.policy_configuration import StoragePolicy

import utils


class FakeStream(object):
    def __init__(self, module, attr):
        self.stream = StringIO()
        self._module = module
        self._attr = attr
        self._orig_attr = None

    def __enter__(self):
        self._orig_attr = getattr(self._module, self._attr)
        setattr(self._module, self._attr, self.stream)

        return self

    def __exit__(self, exc, value, tb):
        setattr(self._module, self._attr, self._orig_attr)


class Namespace(object):
    def __init__(self, **kwargs):
        for (name, value) in kwargs.iteritems():
            setattr(self, name, value)


class TestStoragePolicyLint(unittest.TestCase):
    def test_lint_fails_on_malformed_file(self):
        config = 'test'

        args = Namespace(
            config=StringIO(config))

        with FakeStream(sys, 'stderr') as stderr:
            rc = cli.storage_policy_lint(args)

            utils.assertRegexpMatches(
                stderr.stream.getvalue(),
                'Parsing error:')

            self.assertNotEqual(0, rc)

    def test_lint_fails_on_invalid_config(self):
        config = '[ring:]'

        args = Namespace(
            config=StringIO(config))

        with FakeStream(sys, 'stderr') as stderr:
            rc = cli.storage_policy_lint(args)

            utils.assertRegexpMatches(
                stderr.stream.getvalue(),
                'Configuration error:')

            self.assertNotEqual(0, rc)

    def test_lint_fails_on_exception(self):
        class Stream(object):
            def readline(self):
                raise IOError('Oops')

        args = Namespace(
            config=Stream())

        with FakeStream(sys, 'stderr') as stderr:
            rc = cli.storage_policy_lint(args)

            utils.assertRegexpMatches(
                stderr.stream.getvalue(),
                'Error: Oops')

            self.assertNotEqual(0, rc)

    def test_lint_succeeds_on_valid_config(self):
        config = ''

        args = Namespace(
            config=StringIO(config))

        rc = cli.storage_policy_lint(args)

        self.assertEquals(0, rc)


class TestStoragePolicyQuery(unittest.TestCase):
    def test_load_fails(self):
        config = '[ring:]'

        args = Namespace(
            config=StringIO(config))

        with FakeStream(sys, 'stderr') as stderr:
            rc = cli.storage_policy_query(args)

            utils.assertRegexpMatches(
                stderr.stream.getvalue(),
                'Error: Invalid section name')

            self.assertNotEqual(0, rc)

    def test_lookup_fails(self):
        config = ''

        args = Namespace(
            config=StringIO(config),
            policy_index=1)

        with FakeStream(sys, 'stderr') as stderr:
            rc = cli.storage_policy_query(args)

            utils.assertRegexpMatches(
                stderr.stream.getvalue(),
                'Error: Unknown policy index')

            self.assertNotEqual(0, rc)

    def test_success(self):
        config = '\n'.join(s.strip() for s in '''
            [ring:paris]
            location = paris
            sproxyd_endpoints = http://paris1.int/, http://paris2.int

            [ring:sfo]
            location = sfo
            sproxyd_endpoints = http://sfo1.int

            [storage-policy:2]
            read = sfo
            write = paris
            '''.splitlines())

        args = Namespace(
            config=StringIO(config),
            policy_index=2,
            action=StoragePolicy.WRITE,
            locations=['paris'])

        with FakeStream(sys, 'stdout') as stdout:
            rc = cli.storage_policy_query(args)

            self.assertEqual(0, rc)

            out = stdout.stream.getvalue()

            self.assertTrue('http://paris1.int' in out)
            self.assertTrue('http://paris2.int' in out)

            self.assertFalse('sfo' in out)


class TestMain(unittest.TestCase):
    def test_main(self):
        def exit(code):
            raise SystemExit(code)

        orig_exit = sys.exit
        sys.exit = exit

        # Force failure even when `argparse` is installed on Python 2.6 setups
        orig_argparse = cli.argparse
        if sys.version_info < (2, 7):
            cli.argparse = None

        try:
            with FakeStream(sys, 'stdout') as stdout:
                with FakeStream(sys, 'stderr') as stderr:
                    self.assertRaises(
                        SystemExit,
                        cli.main, ['--help'])

                    if cli.argparse:
                        utils.assertRegexpMatches(
                            stdout.stream.getvalue(),
                            'storage-policy-lint')
                    else:
                        self.assertTrue(sys.version_info < (2, 7))
                        utils.assertRegexpMatches(
                            stderr.stream.getvalue(),
                            'Python 2.7')
        finally:
            sys.exit = orig_exit
            cli.argparse = orig_argparse



import os
import os.path

import fabric.contrib.files

from fabric.api import sudo

from utils import build_object_ring, render


def disk_setup(swift_user):
    # Setup a loopdevice to act as disk for swift
    sudo('mkdir -p /srv')
    sudo('truncate -s 1GB /srv/swift-disk')
    sudo('mkfs.xfs /srv/swift-disk')
    fabric.contrib.files.append(
        filename='/etc/fstab',
        text='/srv/swift-disk /mnt/sdb1 xfs loop,noatime 0 0',
        use_sudo=True
    )

    sudo('mkdir /mnt/sdb1')
    sudo('mount /mnt/sdb1')

    # Prepare directory structure for 4 swift nodes, with two "partitions" each
    node_mkdir = 'mkdir -p /mnt/sdb1/{0:d}/node/sdb{1:d}'
    num_nodes = 4
    for i in range(1, num_nodes + 1):
        sudo(node_mkdir.format(i, i))
        sudo(node_mkdir.format(i, i + num_nodes))
        sudo('ln -s /mnt/sdb1/{0:d} /srv/{1:d}'.format(i, i))
        sudo('mkdir /var/cache/swift{0:d}'.format(i))

    sudo('chown -R {0:s}: /mnt/sdb1'.format(swift_user))
    sudo('mkdir /var/run/swift')
    sudo('chown {0:s}: /var/run/swift /var/cache/swift*'.format(swift_user))
    render(
        directory='assets/saio/phase1/etc',
        filenames=['rc.local'],
        local_path_prefix='assets/saio/phase1',
        content={'user': swift_user},
    )
    sudo('chmod 755 /etc/rc.local')
    sudo('chown root: /etc/rc.local')


def install(swift_user):
    sudo('pip install '
         'git+https://github.com/openstack/python-swiftclient.git@2.6.0')
    sudo('pip install git+https://github.com/openstack/swift.git@2.5.0')

    content = {
        'user': swift_user,
        'group': swift_user,
    }
    for path, _, filenames in os.walk('assets/saio/phase1/etc/swift'):
        render(path, filenames, 'assets/saio/phase1', content)

    sudo('chown -R {0:s}: /etc/swift'.format(swift_user))


def build_rings(swift_user):
    # Account ring
    build_object_ring(
        swift_user=swift_user,
        name='account.builder',
        devices=[
            'r1z1-127.0.0.1:6012/sdb1',
            'r1z2-127.0.0.1:6022/sdb2',
            'r1z3-127.0.0.1:6032/sdb3',
            'r1z4-127.0.0.1:6042/sdb4',
        ],
    )

    # Container ring
    build_object_ring(
        swift_user=swift_user,
        name='container.builder',
        devices=[
            'r1z1-127.0.0.1:6011/sdb1',
            'r1z2-127.0.0.1:6021/sdb2',
            'r1z3-127.0.0.1:6031/sdb3',
            'r1z4-127.0.0.1:6041/sdb4',
        ],
    )

    # Object ring
    build_object_ring(
        swift_user=swift_user,
        name='object.builder',
        devices=[
            'r1z1-127.0.0.1:6010/sdb1',
            'r1z1-127.0.0.1:6010/sdb5',
            'r1z2-127.0.0.1:6020/sdb2',
            'r1z2-127.0.0.1:6020/sdb6',
            'r1z3-127.0.0.1:6030/sdb3',
            'r1z3-127.0.0.1:6030/sdb7',
            'r1z4-127.0.0.1:6040/sdb4',
            'r1z4-127.0.0.1:6040/sdb8',
        ],
    )


def setup_rsync(swift_user):
    render(
        directory='assets/saio/phase1/etc',
        filenames=['rsyncd.conf'],
        local_path_prefix='assets/saio/phase1',
        content={'user': swift_user, 'group': swift_user},
    )
    fabric.contrib.files.sed(
        filename='/etc/default/rsync',
        before='RSYNC_ENABLE=false',
        after='RSYNC_ENABLE=true',
        use_sudo=True,
    )

    sudo('sudo service rsync restart')


def install_scality_swift():
    sudo('pip install '
         'git+https://github.com/scality/scality-sproxyd-client.git')
    sudo('pip install git+https://github.com/scality/ScalitySproxydSwift.git')


def start(swift_user):
    sudo('swift-init main start', user=swift_user)


def stop(swift_user):
    sudo('swift-init main stop', user=swift_user)



import os

import saio
import utils

from fabric.api import env, run, sudo, task
from fabric.context_managers import cd, hide, prefix, settings


@task
def swift(swift_user):
    """
    Bootstrap a SAIO installation.

    :param swift_user: the user to run swift as
    :type swift_user: string
    """
    packages = [
        'curl', 'gcc', 'git-core', 'memcached', 'rsync', 'sqlite3',
        'xfsprogs', 'git-core', 'libffi-dev', 'python-coverage', 'python-dev',
        'python-simplejson', 'python-xattr', 'python-eventlet',
        'python-greenlet', 'python-pastedeploy', 'python-netifaces',
        'python-pip', 'python-dnspython',
    ]

    utils.apt_get(packages)

    # Add swift user
    sudo('useradd -r -s /usr/bin/nologin -d /srv -U {0:s}'.format(swift_user))

    saio.disk_setup(swift_user)
    saio.install(swift_user)
    saio.setup_rsync(swift_user)
    saio.build_rings(swift_user)
    saio.start(swift_user)


@task
def ring():
    """
    Bootstrap Scality RING (environment variable `SCAL_PASS` must be exported).

    The environment variable `SCAL_PASS` is expected to hold username:password
    for fetching scality packages.
    """
    install_env = {
        'SUP_ADMIN_LOGIN': 'supadmin',
        'SUP_ADMIN_PASS': 'supadmin',
        'INTERNAL_MGMT_LOGIN': 'admin',
        'INTERNAL_MGMT_PASS': 'admin',
        'HOST_IP': env.host,
        'SCAL_PASS': os.environ['SCAL_PASS'],
    }
    export_vars = ('{0:s}={1:s}'.format(k, v) for k, v in install_env.items())
    export_cmd = 'export {0:s}'.format(' '.join(export_vars))

    utils.apt_get(['git-core'])
    run('git clone https://github.com/scality/openstack-ci-scripts.git')

    # Hide aborts to not leak any repository passwords to console on failure.
    with cd('openstack-ci-scripts/jenkins'), prefix(export_cmd):
        with prefix('source ring-install.sh'), settings(hide('aborts')):
            run('add_source')
            run('install_base_scality_node', pty=False)  # avoid setup screen
            run('install_supervisor')
            run('install_ringsh')
            run('build_ring')
            run('install_sproxyd')
            run('test_sproxyd')


@task
def scality_storage_policy(swift_user, sproxyd_endpoint):
    """
    Install a storage policy backed by Scality RING.

    :param swift_user: the user swift is running as
    :type swift_user: string
    :param sproxyd_endpoint: the sproxyd endpoint backing the storage policy
    :type sproxyd_endpoint: string
    """
    # Validate sproxyd endpoint connectivity.
    curl_cmd = run('curl --fail --connect-timeout 15 {0:s}/.conf'.format(
                   sproxyd_endpoint), warn_only=True)

    if curl_cmd.failed:
        raise Exception("Unable to establish a connection to sproxyd at "
                        "'{0:s}.'".format(sproxyd_endpoint))

    # Install dependencies.
    saio.install_scality_swift()

    # Add scality RING storage policy.
    content = {
        'sproxyd_endpoint': sproxyd_endpoint,
        'user': swift_user,
    }
    for path, _, filenames in os.walk('assets/saio/phase2/etc/swift'):
        utils.render(path, filenames, 'assets/saio/phase2', content)
    sudo('chown -R {0:s} /etc/swift'.format(swift_user))

    # Setup an additional object ring.
    sudo('mkdir -p /srv/scality/placeholder')
    sudo('chown -R {0:s}: /srv/scality'.format(swift_user))
    sudo('mkdir -p /var/cache/scality')
    sudo('chown {0:s}: /var/cache/scality'.format(swift_user))

    utils.render(
        directory='assets/saio/phase2/etc',
        filenames=['rc.local'],
        local_path_prefix='assets/saio/phase2',
        content={'user': swift_user},
    )
    sudo('chmod 755 /etc/rc.local')
    sudo('chown root: /etc/rc.local')

    utils.build_object_ring(
        swift_user=swift_user,
        name='object-1.builder',
        devices=[
            'r1z1-127.0.0.1:6050/placeholder',
        ],
        replicas=1,
    )

    saio.stop(swift_user)
    saio.start(swift_user)



# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.utils`'''

import unittest

from swift_scality_backend.utils import split_list


class TestSplitList(unittest.TestCase):
    def test_empty_string(self):
        self.assertEqual(
            [],
            list(split_list('')))

    def test_basic(self):
        self.assertEqual(
            ['1', '2', '3'],
            list(split_list('1, 2, 3')))

    def test_space_prefix(self):
        self.assertEqual(
            ['1', '2'],
            list(split_list('   1, 2')))

    def test_space_suffix(self):
        self.assertEqual(
            ['1', '2'],
            list(split_list('1, 2   ')))

    def test_words(self):
        self.assertEqual(
            ['one', 'two', 'three'],
            list(split_list(' one, two, three ')))



# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.http_utils`'''

import contextlib
import os
import socket
import sys
import unittest

import eventlet
import mock

import swift_scality_backend.http_utils
from . import utils


class TestSomewhatBufferedFileObject(unittest.TestCase):
    @contextlib.contextmanager
    def _make_socket(self, addr=None, buffsize=8):
        if not addr:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        else:
            sock = eventlet.connect(addr)

        try:
            yield swift_scality_backend.http_utils.SomewhatBufferedFileObject(
                sock._sock, 'rb', buffsize)
        finally:
            sock.close()

    def test_flush(self):
        with self._make_socket() as sock:
            try:
                ret = sock.flush()
            except NotImplementedError:
                self.fail("Socket flush should not raised a NotImplementedError")
            self.assertTrue(ret is None)

    def test_fail_write(self):
        with self._make_socket() as sock:
            self.assertRaises(NotImplementedError, sock.write, 'abc')

    def test_fail_writelines(self):
        with self._make_socket() as sock:
            self.assertRaises(NotImplementedError, sock.writelines, [])

    def test_fail_call_get_buffered_twice(self):
        with self._make_socket() as sock:
            sock.get_buffered()
            self.assertRaises(
                swift_scality_backend.http_utils.InvariantViolation,
                sock.get_buffered)

    def test_fail_after_get_buffered_read(self):
        with self._make_socket() as sock:
            sock.get_buffered()
            self.assertRaises(
                swift_scality_backend.http_utils.InvariantViolation,
                sock.read)

    def test_fail_after_get_buffered_readline(self):
        with self._make_socket() as sock:
            sock.get_buffered()
            self.assertRaises(
                swift_scality_backend.http_utils.InvariantViolation,
                sock.readline)

    def test_fail_after_get_buffered_readlines(self):
        with self._make_socket() as sock:
            sock.get_buffered()
            self.assertRaises(
                swift_scality_backend.http_utils.InvariantViolation,
                sock.readlines)

    def test_fail_after_get_buffered_next(self):
        with self._make_socket() as sock:
            sock.get_buffered()
            self.assertRaises(
                swift_scality_backend.http_utils.InvariantViolation,
                sock.next)

    def test_interaction(self):
        server_socket = eventlet.listen(('127.0.0.1', 0))

        data = 'abcdefgh\n' * 10

        def server(client, _):
            client.sendall(data)
            client.close()

        server_thread = eventlet.spawn(
            eventlet.serve, server_socket, server)

        try:
            with self._make_socket(
                    server_socket.getsockname()) as sock:
                fst = sock.readline()
                self.assertEqual(fst, 'abcdefgh\n')
                snd = sock.read(size=3)
                self.assertEqual(snd, 'abc')
                buf = sock.get_buffered()
                self.assertEqual(buf, 'defg')

                rest = [buf]
                while True:
                    s = os.read(sock.fileno(), 32)
                    if len(s) == 0:
                        break
                    rest.append(s)

                off = len(fst) + len(snd)
                self.assertEqual(''.join(rest), data[off:])
        finally:
            server_thread.kill()


class TestSomewhatBufferedHTTPConnection(unittest.TestCase):

    @mock.patch('httplib.HTTPResponse.__init__')
    def test_discard_buffering_arg_on_py26(
            self, mock_http_response_init):
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)) \
                as sock:

            kwargs = dict(
                sock=sock, debuglevel=0, strict=0, method=None, buffering=False)

            response_class = \
                swift_scality_backend.http_utils.SomewhatBufferedHTTPConnection.HTTPResponse
            response_obj = response_class(**kwargs)

            if sys.version_info < (2, 7):
                del kwargs['buffering']

            mock_http_response_init.assert_called_once_with(
                response_obj, **kwargs)

    def test_fileno_attr(self):
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)) \
                as sock:
            response_class = \
                swift_scality_backend.http_utils.SomewhatBufferedHTTPConnection.HTTPResponse
            response_obj = response_class(sock=sock)
            self.assertEquals(sock.fileno(), response_obj.fileno())

    @mock.patch('socket._socketobject.makefile')
    def test_makefile_not_called(self, mock_makefile):
        response_class = \
            swift_scality_backend.http_utils.SomewhatBufferedHTTPConnection.HTTPResponse

        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)) \
                as sock:
            response_class(sock=sock, debuglevel=0, strict=0, method=None)

        self.assertFalse(mock_makefile.called)

    def test_non_keepalive_conn(self):
        # https://github.com/scality/ScalitySproxydSwift/issues/93

        content = 'Hello, World!'

        def my_app(env, start_response):
            start_response('200 OK', [('Content-Type', 'text/plain'),
                                      ('Connection', 'close')])
            return [content]

        with utils.WSGIServer(my_app) as server:
            conn_class = \
                swift_scality_backend.http_utils.SomewhatBufferedHTTPConnection
            conn = conn_class(server.ip, server.port)
            conn.request("GET", "/")
            conn_sock_fd = conn.sock.fileno()
            response = conn.getresponse()
            self.assertEqual(conn_sock_fd, response.fileno())



# Copyright (c) 2014, 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''HTTP client utilities'''

import httplib
import operator
import socket
import sys

import eventlet.greenio

from scality_sproxyd_client.exceptions import InvariantViolation
from scality_sproxyd_client.exceptions import SproxydHTTPException


class SomewhatBufferedFileObject(socket._fileobject):
    '''A 'somewhat buffered' file-like object

    This works similar to `socket._fileobject`, which is what you get when
    calling `socket.socket.makefile`, but this one has a couple of differences:

    - All `write`-related methods are removed (raise `NotImplementedError`)
    - It's possible to retrieve the content of the buffer *once*
    - Once the content of the backing buffer has been retrieved, any
      `read`-related method will fail (raise `InvariantViolation`)
    '''

    def get_buffered(self):
        '''Retrieve the buffered data'''

        if self._rbuf is None:
            raise InvariantViolation('Using `get_buffered` more than once')

        value = self._rbuf.getvalue()
        self._rbuf = None
        return value

    def flush(self):
        # We can't raise NotImplementedError here because when an HTTPResponse
        # object is closed, flush() is called
        pass

    def write(self, data):
        raise NotImplementedError

    def writelines(self, lines):
        raise NotImplementedError

    def read(self, size=-1):
        if self._rbuf is None:
            raise InvariantViolation('Using `read` after `get_buffered`')

        return socket._fileobject.read(self, size)

    def readline(self, size=-1):
        if self._rbuf is None:
            raise InvariantViolation('Using `readline` after `get_buffered`')

        return socket._fileobject.readline(self, size)

    def readlines(self, sizehint=0):
        if self._rbuf is None:
            raise InvariantViolation('Using `readllines` after `get_buffered`')

        return socket._fileobject.readlines(self, sizehint)

    def next(self):
        if self._rbuf is None:
            raise InvariantViolation('Using `next` after `get_buffered`')

        return socket._fileobject.next(self)


class SomewhatBufferedHTTPConnection(httplib.HTTPConnection):
    '''A somewhat buffered HTTP connection

    The response type used by this class wraps the underlying socket in a
    `SomewhatBufferedFileObject`.
    '''

    class HTTPResponse(httplib.HTTPResponse):
        '''Like `httplib.HTTPResponse, but with its `fp` attribute wrapped in a
        `SomewhatBufferedFileObject`
        '''
        def __init__(self, sock, debuglevel=0, strict=0, method=None,
                     buffering=False):
            init_args = {
                'sock': sock,
                'debuglevel': debuglevel,
                'strict': strict,
                'method': method,
            }
            if sys.version_info >= (2, 7):
                init_args.update({
                    'buffering': buffering,
                })

            # Prevent makefile calling by stdlib
            makefile = socket._socketobject.makefile
            socket._socketobject.makefile = lambda *_, **_2: None
            try:
                httplib.HTTPResponse.__init__(self, **init_args)
            finally:
                socket._socketobject.makefile = makefile
            # Handle eventlet subtlety
            # https://github.com/eventlet/eventlet/blob/master/eventlet/greenio/base.py#L295
            if isinstance(sock, eventlet.greenio.GreenSocket):
                real_sock = sock.dup()
            else:
                real_sock = sock._sock

            # Fetching in chunks of 1024 bytes seems like a sensible value,
            # since we want to retrieve as little more than the HTTP headers as
            # possible.
            self.fp = SomewhatBufferedFileObject(real_sock, 'rb', 1024)

            # Eventlet finalizing
            # https://github.com/eventlet/eventlet/blob/master/eventlet/greenio/base.py#L299
            if hasattr(real_sock, '_drop') and isinstance(real_sock, eventlet.greenio.GreenSocket):
                real_sock._drop()

        if not hasattr(httplib.HTTPResponse, 'fileno'):
            # py26 compat
            def fileno(self):
                return self.fp.fileno()

    response_class = HTTPResponse

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()


class NoClientAvailable(RuntimeError):
    '''Exception raised when no client with alive endpoints is available'''


class ClientCollection(object):
    '''A collection of back-end connections'''

    def __init__(self, read_clients, write_clients):
        # It's more list rather than tuple, but there's no persistent list in
        # Python
        self._read_clients = tuple(read_clients)
        self._write_clients = tuple(write_clients)

    read_clients = property(operator.attrgetter('_read_clients'))
    write_clients = property(operator.attrgetter('_write_clients'))

    def __repr__(self):
        return 'ClientCollection(read_clients=%r, write_clients=%r)' % \
            (self.read_clients, self.write_clients)

    def __eq__(self, other):
        if isinstance(other, ClientCollection):
            return all([
                self.read_clients == other.read_clients,
                self.write_clients == other.write_clients,
            ])
        else:
            return NotImplemented

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __hash__(self):
        return hash((self.read_clients, self.write_clients))

    @staticmethod
    def _get_client(clients):
        '''Retrieve a client with alive endpoints from a collection of clients

        :param clients: Client collection
        :type clients: iterable of `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :return: A client with alive endpoints
        :rtype: `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :raise NoClientAvailable: No client with available endpoints found
        '''

        for client in clients:
            if client.has_alive_endpoints:
                return client

        raise NoClientAvailable('No alive endpoints available')

    def get_read_client(self):
        '''Return a client at which read operations can be performed

        :return: A client usable for read operations
        :rtype: `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :raise NoClientAvailable: No client with available endpoints found
        '''

        return self._get_client(self.read_clients)

    def get_write_client(self):
        '''Return a client at which write operations can be performed

        :return: A client usable for write operations
        :rtype: `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :raise NoClientAvailable: No client with available endpoints found
        '''

        return self._get_client(self.write_clients)

    def try_read(self, fn):
        '''Attempt a read operation and fallback to write endpoints on 404

        This utility method attempts to retrieve a read endpoint client, and
        call the given function. If it raises an `SproxydHTTPException` with
        `http_status` 404, it falls back to looking up a write endpoint and
        retry the action.

        Other exceptions are passed through.

        :param fn: Callable to attempt the operation
        :type fn: `callable` which takes a `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :return: Result of `fn`

        :raise NoClientAvailable: No client with alive endpoints available
        '''

        try:
            client = self.get_read_client()
        except NoClientAvailable:
            pass
        else:
            try:
                return fn(client)
            except SproxydHTTPException as exc:
                if exc.http_status == 404:
                    pass
                else:
                    raise

        client = self.get_write_client()
        return fn(client)



# Copyright (c) 2014, 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for swift_scality_backend.diskfile"""

import hashlib
import httplib
import logging
import StringIO
import time
import unittest

import mock
import swift.common.exceptions
import swift.common.utils

from swift_scality_backend.diskfile import DiskFileWriter, \
    DiskFileReader, DiskFile, DiskFileManager
from scality_sproxyd_client.exceptions import SproxydHTTPException
from swift_scality_backend.http_utils import ClientCollection
from scality_sproxyd_client.sproxyd_client import SproxydClient
from . import utils
from .utils import make_client_collection


NEW_SPLICE = 'new_splice'
OLD_SPLICE = 'old_splice'
NO_SPLICE_AT_ALL = 'no_splice_at_all'
try:
    import swift.common.splice
    SPLICE = NEW_SPLICE
except ImportError:
    if hasattr(swift.common.utils, 'system_has_splice'):
        SPLICE = OLD_SPLICE
    else:
        SPLICE = NO_SPLICE_AT_ALL


class FakeHTTPResp(httplib.HTTPResponse):

    def __init__(self, status=200):
        self.status = status
        self.reason = 'because'

    def read(self):
        return 'My mock msg'


class FakeHTTPConn(mock.Mock):

    def __init__(self, *args, **kwargs):
        super(FakeHTTPConn, self).__init__(*args, **kwargs)
        self.resp_status = kwargs.get('resp_status', 200)
        self._buffer = StringIO.StringIO()

    def getresponse(self):
        return FakeHTTPResp(self.resp_status)

    def send(self, data):
        self._buffer.write(data)


class TestDiskFileManager(unittest.TestCase):
    """Tests for swift_scality_backend.diskfile.DiskFileManager"""

    def test_init_with_default_splice(self):
        dfm = DiskFileManager({}, mock.Mock())
        self.assertFalse(dfm.use_splice)

    def test_init_with_splice_no(self):
        dfm = DiskFileManager({'splice': 'no'}, mock.Mock())
        self.assertFalse(dfm.use_splice)

    def _test_init_splice_unavailable(self):
        dfm = DiskFileManager({'splice': 'no'}, mock.Mock())
        self.assertFalse(dfm.use_splice, "Splice not wanted by conf and not " +
                         "available from system: use_splice should be False")

        mock_logger = mock.Mock()
        dfm = DiskFileManager({'splice': 'yes'}, mock_logger)
        self.assertFalse(dfm.use_splice, "Splice wanted by conf but not " +
                         "available from system: use_splice should be False")
        self.assertTrue(mock_logger.warn.called)

    def _test_init_splice_available(self):
        dfm = DiskFileManager({'splice': 'yes'}, mock.Mock())
        self.assertTrue(dfm.use_splice, "Splice wanted by conf and " +
                        "available from system: use_splice should be True")

        dfm = DiskFileManager({'splice': 'no'}, mock.Mock())
        self.assertFalse(dfm.use_splice, "Splice not wanted by conf though " +
                         "available from system: use_splice should be False")

    @utils.skipIf(SPLICE != NEW_SPLICE, 'Need new `splice` support')
    @mock.patch('swift.common.splice.splice')
    def test_init_new_splice_unavailable(self, mock_splice):
        type(mock_splice).available = mock.PropertyMock(return_value=False)
        self._test_init_splice_unavailable()

    @utils.skipIf(SPLICE != NEW_SPLICE, 'Need new `splice` support')
    @mock.patch('swift.common.splice.splice')
    def test_init_new_splice_is_available(self, mock_splice):
        type(mock_splice).available = mock.PropertyMock(return_value=True)
        self._test_init_splice_available()

    @utils.skipIf(SPLICE != OLD_SPLICE, 'Need old `splice` support')
    @mock.patch.object(swift.common.utils, 'system_has_splice',
                       return_value=True)
    def test_init_old_splice_is_available(self, mock_splice):
        self._test_init_splice_available()

    @utils.skipIf(SPLICE != OLD_SPLICE, 'Need old `splice` support')
    @mock.patch.object(swift.common.utils, 'system_has_splice',
                       return_value=False)
    def test_init_old_splice_unavailable(self, mock_splice):
        self._test_init_splice_unavailable()

    @utils.skipIf(SPLICE != NO_SPLICE_AT_ALL, 'This Swift knows `splice`')
    def test_init_no_splice_at_all(self):
        self._test_init_splice_unavailable()

    def test_get_diskfile(self):
        client_collection = make_client_collection()
        dfm = DiskFileManager({}, mock.Mock())

        diskfile = dfm.get_diskfile(client_collection, 'a', 'c', 'o')
        self.assertTrue(isinstance(diskfile, DiskFile))


class TestDiskFileWriter(unittest.TestCase):
    """Tests for swift_scality_backend.diskfile.DiskFileWriter"""

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(None, None))
    def test_init(self, mock_http):
        client_collection = make_client_collection()
        # Note the white space, to test proper URL encoding
        DiskFileWriter(client_collection, 'ob j', logger=logging.root)

        expected_header = {'transfer-encoding': 'chunked'}
        mock_http.assert_called_once_with('ob j', expected_header)

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(FakeHTTPConn(resp_status=404), None))
    def test_put_with_404_response(self, mock_http):
        client_collection = make_client_collection()
        dfw = DiskFileWriter(client_collection, 'obj', logger=logging.root)

        fake_http_conn = mock_http.return_value[0]
        msg = r'.*404 / %s.*' % fake_http_conn.getresponse().read()
        utils.assertRaisesRegexp(SproxydHTTPException, msg, dfw.put, {})

        fake_http_conn.close.assert_called_once_with()

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(FakeHTTPConn(), mock.Mock()))
    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.put_meta')
    def test_put_with_200_response(self, mock_put_meta, mock_http):
        client_collection = make_client_collection()
        dfw = DiskFileWriter(client_collection, 'obj', logger=logging.root)

        dfw.put({'meta1': 'val'})

        fake_http_conn = mock_http.return_value[0]
        self.assertEqual('0\r\n\r\n', fake_http_conn._buffer.getvalue())

        mock_release_conn = mock_http.return_value[1]
        mock_release_conn.assert_called_once_with()

        mock_put_meta.assert_called_with('obj', {
            'df': {'meta1': 'val'},
            'meta1': 'val',
            'mf': {},
            'ETag': 'd41d8cd98f00b204e9800998ecf8427e',
            'name': 'obj'
        })

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(FakeHTTPConn(), mock.Mock()))
    def test_write_no_data(self, mock_http):
        dfw = DiskFileWriter(make_client_collection(), 'obj',
                             logger=logging.root)

        written = dfw.write("")

        self.assertEqual(0, written)
        fake_http_conn = mock_http.return_value[0]
        self.assertEqual('0\r\n\r\n', fake_http_conn._buffer.getvalue())

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(FakeHTTPConn(), mock.Mock()))
    def test_write_with_data(self, mock_http):
        dfw = DiskFileWriter(make_client_collection(), 'obj',
                             logger=logging.root)

        data = "a" * 4096
        written = dfw.write(data)

        self.assertEqual(len(data), written)
        fake_http_conn = mock_http.return_value[0]
        self.assertEqual('%x\r\n%s\r\n' % (len(data), data),
                         fake_http_conn._buffer.getvalue())


class TestDiskFileReader(unittest.TestCase):

    def test_app_iter_ranges_with_no_ranges(self):
        dfr = DiskFileReader(make_client_collection(), 'obj', False,
                             logger=logging.root)

        gen = dfr.app_iter_ranges([], mock.sentinel.arg1,
                                  mock.sentinel.arg2, mock.sentinel.arg3)
        self.assertEqual('', gen.next())
        self.assertRaises(StopIteration, gen.next)

    @mock.patch('swift.common.swob.multi_range_iterator')
    def test_app_iter_ranges(self, mock_mri):
        mock_mri.return_value = iter(['data'])
        dfr = DiskFileReader(make_client_collection(), 'obj', False,
                             logger=logging.root)

        gen = dfr.app_iter_ranges([(1, 100)], mock.sentinel.arg1,
                                  mock.sentinel.arg2, mock.sentinel.arg3)

        self.assertEqual('data', gen.next())
        self.assertRaises(StopIteration, gen.next)
        mock_mri.assert_called_once_with([(1, 100)], mock.sentinel.arg1,
                                         mock.sentinel.arg2, mock.sentinel.arg3,
                                         dfr.app_iter_range)


class TestDiskFile(unittest.TestCase):
    """Tests for swift_scality_backend.diskfile.DiskFile"""

    @staticmethod
    def hash_str(strings):
        "Get a SHA1 hex hash of the iterable of `str` passed as arguments"

        sha1 = hashlib.sha1()
        for string in strings:
            sha1.update(string)
        return sha1.hexdigest()

    def test_init_quotes_object_path(self):
        acc, cont, obj = 'a', '@/', '/ob/j'

        sproxyd_client = SproxydClient(['http://host:81/path/'], logger=mock.Mock())
        df = DiskFile(sproxyd_client, acc, cont, obj,
                      use_splice=False, logger=logging.root)
        self.assertEqual(self.hash_str([acc, cont, obj]), df._name)

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_meta',
                return_value=None)
    def test_open_when_no_metadata(self, mock_get_meta):
        acc, cont, obj = 'a', 'c', 'o'

        client_collection = make_client_collection()
        df = DiskFile(client_collection, acc, cont, obj, use_splice=False,
                      logger=logging.root)

        self.assertRaises(swift.common.exceptions.DiskFileDeleted, df.open)
        mock_get_meta.assert_called_once_with(self.hash_str([acc, cont, obj]))

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_meta',
                return_value={'name': 'o'})
    def test_open(self, mock_get_meta):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        df.open()

        self.assertEqual({'name': 'o', 'mf': {}, 'df': {'name': 'o'}}, df._metadata)

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_meta')
    def test_open_expired_file(self, mock_get_meta):
        acc, cont, obj = 'a', 'c', 'o'
        mock_get_meta.return_value = {'X-Delete-At': time.time() - 10}

        client_collection = make_client_collection()
        df = DiskFile(client_collection, acc, cont, obj, use_splice=False,
                      logger=logging.root)
        self.assertRaises(swift.common.exceptions.DiskFileExpired, df.open)
        mock_get_meta.assert_called_once_with(self.hash_str([acc, cont, obj]))

    def test_get_metadata_when_diskfile_not_open(self):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        self.assertRaises(swift.common.exceptions.DiskFileNotOpen,
                          df.get_metadata)

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_meta',
                return_value={'name': 'o'})
    def test_read_metadata(self, mock_get_meta):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        metadata = df.read_metadata()

        self.assertEqual({'name': 'o'}, metadata)

    def test_reader(self):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        reader = df.reader()
        self.assertTrue(isinstance(reader, DiskFileReader))

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.get_http_conn_for_put',
                return_value=(FakeHTTPConn(), mock.Mock()))
    def test_create(self, mock_http):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        with df.create() as writer:
            self.assertTrue(isinstance(writer, DiskFileWriter))

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.put_meta')
    def test_write_metadata(self, mock_put_meta):
        acc, cont, obj = 'a', 'c', 'o'

        client_collection = make_client_collection()
        df = DiskFile(client_collection, acc, cont, obj, use_splice=False,
                      logger=logging.root)

        df.write_metadata({'k': 'v'})

        mock_put_meta.assert_called_once_with(self.hash_str([acc, cont, obj]), {
            'df': {'name': 'ae541120074a053234ef14e1b050ee3a7433099c'},
            'k': 'v',
            'mf': {'k': 'v'}
        })

    @mock.patch('scality_sproxyd_client.sproxyd_client.SproxydClient.del_object')
    def test_delete(self, mock_del_object):
        acc, cont, obj = 'a', 'c', 'o'

        client_collection = make_client_collection()
        df = DiskFile(client_collection, acc, cont, obj, use_splice=False,
                      logger=logging.root)

        df.delete("ignored")

        mock_del_object.assert_called_once_with(self.hash_str([acc, cont, obj]))

    @utils.skipIf(not hasattr(swift.common.utils, 'Timestamp'), 'Swift2+ only')
    def test_timestamps_when_no_metadata(self):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        # assertRaises expects a `callable`, but `timestamp` is a property
        # We could use the `with self.assertRaises(exc):` form but that's
        # Python 2.7+ only
        self.assertRaises(swift.common.exceptions.DiskFileNotOpen,
                          lambda: df.timestamp)
        self.assertRaises(swift.common.exceptions.DiskFileNotOpen,
                          lambda: df.data_timestamp)

    @utils.skipIf(not hasattr(swift.common.utils, 'Timestamp'), 'Swift2+ only')
    @mock.patch('swift.common.utils.Timestamp')
    def test_timestamp(self, mock_timestamp):
        client_collection = make_client_collection()
        df = DiskFile(client_collection, 'a', 'c', 'o', use_splice=False,
                      logger=logging.root)

        df._metadata = mock.Mock()
        df.timestamp

        df._metadata.get.assert_called_once_with('X-Timestamp')
        mock_timestamp.assert_called_once_with(df._metadata.get())


class TestClientCollection(unittest.TestCase):
    '''Tests for `swift_scality_backend.diskfile.ClientCollection`'''

    def test_constructor(self):
        def make_iter():
            cell = [False]

            def iter():
                yield ()
                cell[0] = True
                yield ()

            return cell, iter()

        read_set_cell, read_set = make_iter()
        write_set_cell, write_set = make_iter()

        ClientCollection(read_set, write_set)

        self.assertTrue(read_set_cell[0])
        self.assertTrue(write_set_cell[0])

    def test_hash(self):
        col = ClientCollection([None, None], [None])
        hash(col)



import hashlib
import json
import io
import os

import swiftclient.client

from fabric.api import task
from swiftclient.exceptions import ClientException

AUTH_ENDPOINT = 'http://{host:s}:8080/auth/v1.0'


def hash_dataset(directory, manifest_path=None):
    """
    SHA1 hash the content of each file in the given directory.

    :param directory: containing directory of files to compute hashes of
    :type directory: string
    :param manifest_path: path to where a manifest of all files and
        corresponding SHA1 hash will be written
    :type manifest_path: string
    :return: dictionary mapping filename to a hash of its content
    """
    file_hashes = {}
    for filename in os.listdir(directory):
        path = os.path.join(directory, filename)
        if os.path.isfile(path) and os.access(path, os.R_OK):
            with io.open(path, 'rb') as f:
                file_hashes[path] = hashlib.sha1(f.read()).hexdigest()
        else:
            print("Skipped '{0:s}'".format(path))

    manifest = manifest_path or os.path.join(directory, 'manifest')
    with io.open(manifest, 'wb') as manifest_file:
        json.dump(file_hashes, manifest_file, indent=2)

    return file_hashes


def summarize(failures):
    """
    Print a summary of any failures to stdout.

    :param failues: mapping of objects with bad integrity to an error string
    :type failures: dict
    """
    if failures:
        print("Objects with bad integrity encountered:")
        for obj, reason in failures.items():
            print("- {obj:<32s} {reason:s}".format(obj=obj, reason=reason))
    else:
        print("All objects checked out OK")


def create_container(swift_connection, name, storage_policy=None):
    """
    Create a container with an optional storage policy.

    :param swift_connection: connection to Swift
    :type swift_connection: :py:class:`swiftclient.client.Connection`
    :param name: container name
    :type name: string
    :param storage_policy: container storage policy (optional)
    :type storage_policy: string
    :return: policy of the craeted container
    """
    headers = {'X-Storage-Policy': storage_policy} if storage_policy else {}
    swift_connection.put_container(name, headers)
    return swift_connection.head_container(name).get('x-storage-policy')


def upload(swift_connection, container, files):
    """
    Upload a set of files to a container.

    :param swift_connection: connection to Swift
    :type swift_connection: :py:class:`swiftclient.client.Connection`
    :param container: target container for upload
    :type container: string
    :param files: paths to files for upload
    :type files: list of strings
    """
    for filename in files:
        with io.open(filename, 'rb') as f:
            swift_connection.put_object(container, filename, f)


def integrity_check(swift_connection, container, file_hashes):
    """
    Download and verify the integrity of a set of files from a bucket.

    :param swift_connection: connection to Swift
    :type swift_connection: :py:class:`swiftclient.client.Connection`
    :param container: container for integrity check
    :type container: string
    :param file_hashes: dictionary of mapping filename to a SHA1 hash
        of its content
    :type file_hashes: dict
    :return: dict of object names with bad integrity
    """
    failures = {}
    for filename, expected_sha1 in file_hashes.items():
        print("Checking integrity of '{0:s}'...".format(filename))
        try:
            headers, content = swift_connection.get_object(container, filename)
            actual_sha1 = hashlib.sha1(content).hexdigest()
            if expected_sha1 != actual_sha1:
                failures[filename] = (
                    "SHA1 mismatch: {expected:s} != {actual:s}".format(
                        expected=expected_sha1,
                        actual=actual_sha1,
                    )
                )

        except ClientException as e:
            failures[filename] = (
                "Failed to retrieve object '{obj:s}': {exception:s}".format(
                    obj=filename,
                    exception=e,
                )
            )

    return failures


@task
def container_integrity(swift_host, container, directory, storage_policy=None,
                        manifest_path=None, user='test:tester', key='testing'):
    """
    Upload files to a container and validate integrity.

    :param swift_host: hostname or IP of SAIO installation
    :type swift_host: string
    :param container: target container for upload and integrity check
    :type container: string
    :param directory: path to directory holding dataset for upload, and
        integrity check
    :type directory: string
    :param storage_policy: storage policy to apply on created container
    :type storage_policy: string
    :param manifest_path: path to where a manifest of all checked files will
        be written
    :type manifest_path: string
    :param user: username for authentication
    :type user: string
    :param key: password to authenticate with
    :type key: string
    """
    file_hashes = hash_dataset(directory, manifest_path)
    swift_connection = swiftclient.client.Connection(
        authurl=AUTH_ENDPOINT.format(host=swift_host),
        user=user,
        key=key,
    )
    try:
        policy = create_container(swift_connection, container, storage_policy)
        print("'{0:s}' created with policy '{1:s}'".format(container, policy))

        upload(swift_connection, container, file_hashes.keys())
        failures = integrity_check(swift_connection, container, file_hashes)
        summarize(failures)
    finally:
        swift_connection.close()


@task
def container_by_manifest(swift_host, container, manifest_path,
                          user='test:tester', key='testing'):
    """
    Validate the integrity of the files in a container from a manifest.

    :param swift_host: hostname or IP of SAIO installation
    :type swift_host: string
    :param container: container to inspect
    :type container: string
    :param manifest_path: path to the manifest file
    :type manifest_path: string
    :param user: username for authentication
    :type user: string
    :param key: password to authenticate with
    :type key: string
    """
    with io.open(manifest_path, 'rb') as manifest_file:
        manifest = json.load(manifest_file)

    swift_connection = swiftclient.client.Connection(
        authurl=AUTH_ENDPOINT.format(host=swift_host),
        user=user,
        key=key,
    )
    try:
        failures = integrity_check(swift_connection, container, manifest)
        summarize(failures)
    finally:
        swift_connection.close()



# Copyright (c) 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Structures and procedures to handle Swift storage policy configuration files.
'''

import ConfigParser
import operator
import os.path
import urlparse

import swift.common.manager

from swift_scality_backend import utils


DEFAULT_CONFIGURATION_PATH = os.path.join(
    swift.common.manager.SWIFT_DIR, 'scality-storage-policies.ini')


class Endpoint(object):  # pylint: disable=R0903
    '''Representation of an Sproxyd endpoint URL

    This is a restricted version of :class:`urlparse.ParseResult`. The
    :attr:`scheme`, :attr:`netloc` and :attr:`path` attributes of instances of
    this type correspond to the same attributes on
    :class:`urlparse.ParseResult` objects.

    The constructor expects a URL in its `str` format, or an already-parsed URL
    compatible with :class:`urlparse.ParseResult`::

        >>> Endpoint('http://localhost')
        Endpoint(url='http://localhost')
        >>> Endpoint(urlparse.urlparse('http://localhost'))
        Endpoint(url='http://localhost')

    An :class:`Endpoint` can't have `params`, `query` or `fragment` portions in
    its URL. This results in a :exc:`ValueError` at construction time::

        >>> Endpoint('http://localhost/?a=b')
        Traceback (most recent call last):
        ...
        ValueError: Endpoint URL can't have query values

    The string-representation of an :class:`Endpoint` behaves as the URL it
    represents::

        >>> print Endpoint('http://localhost')
        http://localhost

    They have some useful attributes::

        >>> localhost = Endpoint('http://localhost/path')
        >>> localhost.scheme
        'http'
        >>> localhost.netloc
        'localhost'
        >>> localhost.path
        '/path'
        >>> localhost.url # doctest: +ELLIPSIS
        ParseResult(scheme='http', netloc='localhost', path='/path', ...)

    An :class:`Endpoint` can be compared to others, to URL strings or to
    :class:`urlparse.ParseResult` instances::

        >>> Endpoint('http://localhost/path') == 'http://localhost/path'
        True
        >>> Endpoint('http://localhost') != Endpoint('http://otherhost')
        True
        >>> Endpoint('http://localhost') == \
                urlparse.urlparse('http://otherhost')
        False

    :param url: URL of the endpoint
    :type url: `str` or :class:`urlparse.ParseResult`

    :raise ValueError: URL has params
    :raise ValueError: URL has query values
    :raise ValueError: URL has a fragment
    '''

    __slots__ = '_url',

    def __init__(self, url):
        if isinstance(url, basestring):
            url = urlparse.urlparse(url)

        if url.params:
            raise ValueError('Endpoint URL can\'t have params')

        if url.query:
            raise ValueError('Endpoint URL can\'t have query values')

        if url.fragment:
            raise ValueError('Endpoint URL can\'t have a fragment')

        self._url = url

    url = property(operator.attrgetter('_url'), doc='Parsed URL')

    def __str__(self):
        return urlparse.urlunparse(self.url)

    def __repr__(self):
        return 'Endpoint(url=%r)' % urlparse.urlunparse(self.url)

    def __eq__(self, other):
        if isinstance(other, basestring):
            return self.url == urlparse.urlparse(other)
        elif isinstance(other, urlparse.ParseResult):
            return self.url == other
        elif isinstance(other, Endpoint):
            return self.url == other.url
        else:
            return NotImplemented

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __hash__(self):
        return hash(self.url)

    scheme = property(operator.attrgetter('url.scheme'),
                      doc='Endpoint URL scheme')
    netloc = property(operator.attrgetter('url.netloc'),
                      doc='Endpoint URL netloc')
    path = property(operator.attrgetter('url.path'), doc='Endpoint URL path')


class Location(object):  # pylint: disable=R0903
    '''Representation of a Ring location

    Values of this type have only one useful attribute, :attr:`name`::

        >>> Location('paris').name
        'paris'

    They can be compared to `str` or another :class:`Location`::

        >>> Location('paris') == 'paris'
        True
        >>> Location('paris') == Location('london')
        False
        >>> Location('paris') != 'sfo'
        True

    :param name: Name of the location
    :type name: `str`
    '''

    __slots__ = '_name',

    def __init__(self, name):
        self._name = name

    name = property(operator.attrgetter('_name'), doc='Location name')

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Location(name=%r)' % self.name

    def __eq__(self, other):
        if isinstance(other, basestring):
            return self.name == other
        elif isinstance(other, Location):
            return self.name == other.name
        else:
            return NotImplemented

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __hash__(self):
        return hash(self.name)


class Ring(object):  # pylint: disable=R0903
    '''Representation of a Ring

    A :class:`Ring` has a name, a location and a set of endpoints, available as
    attributes. The constructor supports `str` values as inputs, but these will
    be turned into :class:`Location` and :class:`Endpoint` objects::

        >>> r = Ring('paris-arc6+3', location='paris', \
                     endpoints=['http://localhost', 'http://otherhost'])
        >>> r # doctest: +ELLIPSIS
        Ring(name='...', location=Location(...), endpoints=frozenset([End...
        >>> r.name, r.location, r.endpoints # doctest: +ELLIPSIS
        ('paris-arc6+3', Location(name='paris'), frozenset([Endpoint(...),...]))

    Furthermore, a :class:`Ring` is iterable (over its endpoints).

        >>> for endpoint in sorted(e.netloc for e in r):
        ...     print endpoint
        localhost
        otherhost
        >>> 'http://localhost' in r
        True
        >>> Endpoint('http://someotherhost') in r
        False

    When `location` is a `basestring`, it will be converted into a
    :class:`Location`.

    When elements of `endpoints` are no :class:`Endpoint` instances, they are
    converted into :class:`Endpoint`.

    :param name: Name of the Ring
    :type name: `str`
    :param location: Location of the Ring
    :type location: :class:`Location` or `str`
    :param endpoints: Ring endpoints
    :type endpoints: Iterable of `str` or :class:`Endpoint`
    '''

    __slots__ = '_name', '_location', '_endpoints',

    def __init__(self, name, location, endpoints):
        self._name = name

        self._location = location if not isinstance(location, basestring) \
            else Location(location)

        self._endpoints = frozenset(
            endpoint if isinstance(endpoint, Endpoint) else Endpoint(endpoint)
            for endpoint in endpoints)

    name = property(operator.attrgetter('_name'), doc='Ring name')
    location = property(operator.attrgetter('_location'), doc='Ring location')
    endpoints = property(operator.attrgetter('_endpoints'),
                         doc='Ring endpoints')

    def __iter__(self):
        return iter(self.endpoints)

    def __repr__(self):
        return 'Ring(name=%r, location=%r, endpoints=%r)' % \
            (self.name, self.location, self.endpoints)

    def __eq__(self, other):
        if not isinstance(other, Ring):
            return NotImplemented

        svals = (self.name, self.location, self.endpoints)
        ovals = (other.name, other.location, other.endpoints)

        return svals == ovals

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __hash__(self):
        return hash((self.name, self.location, self.endpoints))


class StoragePolicy(object):  # pylint: disable=R0903
    '''Representation of a storage-policy

    A :class:`StoragePolicy` has an index, a set of read-only rings and a set
    of writable rings::

        >>> ring1 = Ring('paris-arc6+3', 'paris', ['http://paris/arc6+3'])
        >>> ring2 = Ring('london-arc6+3', 'london', ['http://london/arc6+3'])
        >>> p = StoragePolicy(1, [ring1], [ring2])
        >>> p # doctest: +ELLIPSIS
        StoragePolicy(index=1, read_set=frozenset([Ring(...)]), write_set=...)
        >>> p.index
        1
        >>> p.read_set # doctest: +ELLIPSIS
        frozenset([Ring(name='paris-arc6+3', location=..., endpoints=...)])
        >>> p.write_set # doctest: +ELLIPSIS
        frozenset([Ring(name='london-arc6+3', location=..., endpoints=...)])

    :param index: Policy index
    :type index: `int`
    :param read_set: Set of read-only rings
    :type read_set: Iterable of :class:`Ring`
    :param write_set: Set of writeable rings
    :type write_set: Iterable of :class:`Ring`
    '''

    READ = 'read'
    WRITE = 'write'

    __slots__ = '_index', '_read_set', '_write_set',

    def __init__(self, index, read_set, write_set):
        self._index = index
        self._read_set = frozenset(read_set)
        self._write_set = frozenset(write_set)

    index = property(operator.attrgetter('_index'), doc='Policy index')
    read_set = property(operator.attrgetter('_read_set'), doc='Policy read set')
    write_set = property(operator.attrgetter('_write_set'),
                         doc='Policy write set')

    def __repr__(self):
        return 'StoragePolicy(index=%r, read_set=%r, write_set=%r)' % \
            (self.index, self.read_set, self.write_set)

    def __eq__(self, other):
        if not isinstance(other, StoragePolicy):
            return NotImplemented

        svals = (self.index, self.read_set, self.write_set)
        ovals = (other.index, other.read_set, other.write_set)

        return svals == ovals

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __hash__(self):
        return hash((self.index, self.read_set, self.write_set))

    def lookup(self, method, location_hints=None):
        '''Lookup endpoints for a given method

        This method calculates an (ordered) list of sets of :class:`Endpoint`
        that can be used to perform a certain type of action, based on an
        optional list of location preferences.

        A simple example::

            >>> sp = StoragePolicy(
            ...     index=1,
            ...     read_set=[
            ...         Ring(
            ...             name='sfo-arc6+3',
            ...             location='sfo',
            ...             endpoints=[
            ...                 'http://sfo1.int/arc6+3']),
            ...         Ring(
            ...             name='nyc-arc6+3',
            ...             location='nyc',
            ...             endpoints=[
            ...                 'http://nyc1.int/arc6+3',
            ...                 'http://nyc2.int/arc6+3'])],
            ...     write_set=[
            ...         Ring(
            ...             name='paris-arc6+3',
            ...             location='paris',
            ...             endpoints=[
            ...                 'http://paris1.int/arc6+3',
            ...                 'http://paris2.int/arc6+3'])])

            >>> def dump(result):
            ...     for (idx, endpoints) in enumerate(result):
            ...         print idx, sorted(str(ep) for ep in endpoints)

            >>> dump(sp.lookup(StoragePolicy.READ))
            0 ['http://nyc1.int/arc6+3', 'http://nyc2.int/arc6+3', 'http://paris1.int/arc6+3', 'http://paris2.int/arc6+3', 'http://sfo1.int/arc6+3']

            >>> dump(sp.lookup(StoragePolicy.WRITE))
            0 ['http://paris1.int/arc6+3', 'http://paris2.int/arc6+3']

            >>> dump(sp.lookup(StoragePolicy.READ, ['sfo']))
            0 ['http://sfo1.int/arc6+3']
            1 ['http://nyc1.int/arc6+3', 'http://nyc2.int/arc6+3', 'http://paris1.int/arc6+3', 'http://paris2.int/arc6+3']

            >>> dump(sp.lookup(StoragePolicy.READ, ['sfo', 'paris']))
            0 ['http://sfo1.int/arc6+3']
            1 ['http://paris1.int/arc6+3', 'http://paris2.int/arc6+3']
            2 ['http://nyc1.int/arc6+3', 'http://nyc2.int/arc6+3']

        :param method: Target action (one of :const:`StoragePolicy.READ` or
                        :const:`StoragePolicy.WRITE`)
        :type method: `str`
        :param location_hints: List of preferred locations, in descending order
        :type location_hints: Iterable of `str` or :class:`Location`

        :return: List of target :class:`Endpoint`
        :rtype: Iterable of `frozenset` of :class:`Endpoint`

        :raise ValueError: Invalid `method` passed
        '''

        # Step 1: Gather all candidate target rings
        if method == self.READ:
            candidates0 = self.read_set.union(self.write_set)
        elif method == self.WRITE:
            candidates0 = self.write_set
        else:
            raise ValueError('Invalid method: %r' % method)

        # Step 2: Sort all candidates based on the location hints
        def worker((temp_result, candidates), hint):
            # Step 2a: Find all rings we didn't process yet at current hint
            here = frozenset(
                ring for ring in candidates if ring.location == hint)

            # Step 2b: Update the result, and remove from the unused rings
            return (temp_result + [here], candidates.difference(here))

        location_hints = location_hints or []
        initial_state = ([], candidates0)
        result, leftover_candidates = reduce(worker, location_hints, initial_state)
        result.append(leftover_candidates)

        # Step 3: Turn every set of rings into a set of endpoints, skipping
        # empty sets
        return [frozenset(
            endpoint
                for ring in rings
                for endpoint in ring)
                for rings in result if rings]


RING_SECTION_PREFIX = 'ring:'
STORAGE_POLICY_SECTION_PREFIX = 'storage-policy:'

RING_LOCATION_OPTION = 'location'
RING_SPROXYD_ENDPOINTS_OPTION = 'sproxyd_endpoints'

STORAGE_POLICY_READ_OPTION = 'read'
STORAGE_POLICY_WRITE_OPTION = 'write'


class ConfigurationError(Exception):
    '''Exception thrown when a configuration is somehow invalid.'''


class Configuration(object):
    '''Representation of a storage-policy configuration

    This is a set of :class:`StoragePolicy` objects with utilities to retrieve
    one based on index, read them from a configuration file or write them as a
    new configuration.

        >>> conf = Configuration(
        ...     [StoragePolicy(1, [], []), StoragePolicy(2, [], [])])
        >>> conf # doctest: +ELLIPSIS
        Configuration(policies=[StoragePolicy(index=1, ...])

    A :class:`Configuration` is iterable over its policies::

        >>> print list(sorted(policy.index for policy in conf))
        [1, 2]

    In most application, a policy will be looked up based on its index::

        >>> conf.get_policy(index=2)
        StoragePolicy(index=2, read_set=frozenset([]), write_set=frozenset([]))
        >>> conf.get_policy(index=3)
        Traceback (most recent call last):
        ...
        ValueError: Unknown policy index: 3

    :param policies: Policies contained in the configuration
    :type policies: Iterable of :class:`StoragePolicy`

    :raise ValueError: Duplicate policy index found
    '''

    __slots__ = '_policies_map',

    def __init__(self, policies):
        self._policies_map = {}

        for policy in policies:
            if policy.index in self._policies_map:
                raise ValueError('Duplicate policy index: %r' % policy.index)

            self._policies_map[policy.index] = policy

    def __repr__(self):
        return 'Configuration(policies=%r)' % list(self)

    def __eq__(self, other):
        if not isinstance(other, Configuration):
            return NotImplemented

        return frozenset(self) == frozenset(other)

    def __ne__(self, other):
        equal = self.__eq__(other)

        if equal is NotImplemented:
            return NotImplemented
        else:
            return not equal

    def __iter__(self):
        return self._policies_map.itervalues()

    def __hash__(self):
        return hash(frozenset(self))

    def get_policy(self, index):
        '''Retrieve a policy based on its index

        :param index: Policy to look up
        :type index: Type of :attr:`StoragePolicy.index`

        :raise ValueError: Unknown policy index
        '''

        if index in self._policies_map:
            return self._policies_map[index]
        else:
            raise ValueError('Unknown policy index: %r' % index)

    @classmethod
    def from_stream(cls, stream, filename=None):
        '''Parse a configuration from a stream

        This functions turns a configuration file into a
        :class:`Configuration`, with some sanity checks along the way. It is
        based on :class:`ConfigParser.SafeConfigParser`.

        The `stream` object must have a :meth:`readline` method. `filename`
        will be used in error reporting, if available.

        :param stream: Stream to parse
        :type stream: File-like object
        :param filename: Filename of input
        :type filename: `str`

        :returns: Parsed :class:`Configuration`
        :rtype: :class:`Configuration`

        :raise ConfigurationError: Various configurations issues detected
        '''

        parser = ConfigParser.SafeConfigParser()
        parser.readfp(stream, filename)

        ring_sections = [
            section for section in parser.sections()
            if section.startswith(RING_SECTION_PREFIX)]
        sp_sections = [
            section for section in parser.sections()
            if section.startswith(STORAGE_POLICY_SECTION_PREFIX)]

        _default = object()

        def get(section, setting, default=_default):
            '''Safely retrieve a value from a configuration.'''

            if not parser.has_option(section, setting):
                if default is not _default:
                    return default
                else:
                    raise ConfigurationError(
                        'Section %r lacks a %r setting' % (section, setting))

            return parser.get(section, setting)

        rings = {}
        for section in ring_sections:
            name = section[len(RING_SECTION_PREFIX):]
            if not name:
                raise ConfigurationError('Invalid section name %r' % section)

            location = get(section, RING_LOCATION_OPTION)
            if not location:
                raise ConfigurationError(
                    'Invalid %r setting in %r' %
                    (RING_LOCATION_OPTION, section))

            endpoints = get(section, RING_SPROXYD_ENDPOINTS_OPTION)

            endpoints2 = set()
            for endpoint in utils.split_list(endpoints):
                try:
                    endpoints2.add(Endpoint(endpoint))
                except ValueError as exc:
                    raise ConfigurationError(
                        'Error parsing endpoint %r in %r: %s' %
                        (endpoint, section, exc.message))

            if not endpoints2:
                raise ConfigurationError(
                    'Invalid %r setting in %r' %
                    (RING_SPROXYD_ENDPOINTS_OPTION, section))

            rings[name] = Ring(name, location, endpoints2)

        policies = list()
        for section in sp_sections:
            index = section[len(STORAGE_POLICY_SECTION_PREFIX):]
            if not index:
                raise ConfigurationError('Invalid section name %r' % section)

            try:
                index2 = int(index)
            except (ValueError, TypeError):
                raise ConfigurationError('Invalid policy index: %r' % index)

            read = get(section, STORAGE_POLICY_READ_OPTION, '')
            write = get(section, STORAGE_POLICY_WRITE_OPTION)

            read2 = set()
            for ring in utils.split_list(read):
                if ring not in rings:
                    raise ConfigurationError(
                        'Unknown %r ring %r in policy %r' %
                        (STORAGE_POLICY_READ_OPTION, ring, index2))

                read2.add(rings[ring])

            write2 = list(utils.split_list(write))

            if len(write2) > 1:
                raise ConfigurationError('Multiple %r rings defined in %r' %
                                         (STORAGE_POLICY_WRITE_OPTION, section))

            write3 = set()
            for ring in write2:
                if ring not in rings:
                    raise ConfigurationError(
                        'Unknown %r ring %r in policy %r' %
                        (STORAGE_POLICY_WRITE_OPTION, ring, index2))

                write3.add(rings[ring])

            policies.append(StoragePolicy(index2, read2, write3))

        return cls(policies)

    def to_stream(self, stream):
        '''Unparse a :class:`Configuration` into a stream

        :param stream: Stream to write the :class:`Configuration` to,
                        in INI-format
        :type stream: File-like object
        '''

        rings = set()

        for policy in self:
            for ring in policy.read_set:
                rings.add(ring)

            for ring in policy.write_set:
                rings.add(ring)

        parser = ConfigParser.SafeConfigParser()

        for ring in rings:
            section_name = RING_SECTION_PREFIX + ring.name
            parser.add_section(section_name)

            parser.set(section_name, RING_LOCATION_OPTION, ring.location.name)
            parser.set(
                section_name, RING_SPROXYD_ENDPOINTS_OPTION,
                ', '.join(str(endpoint) for endpoint in ring.endpoints))

        for policy in self:
            section_name = '%s%r' % \
                (STORAGE_POLICY_SECTION_PREFIX, policy.index)
            parser.add_section(section_name)

            parser.set(
                section_name, STORAGE_POLICY_READ_OPTION,
                ', '.join(ring.name for ring in policy.read_set))
            parser.set(
                section_name, STORAGE_POLICY_WRITE_OPTION,
                ', '.join(ring.name for ring in policy.write_set))

        parser.write(stream)



# Copyright (c) 2010-2013 OpenStack, LLC.
# Copyright (c) 2014, 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Sproxyd Disk File Interface for Swift Object Server"""

import contextlib
import hashlib
import httplib
import operator
import time
import urlparse
import copy

import eventlet
import eventlet.green.os
import swift.common.bufferedhttp
import swift.common.exceptions
import swift.common.swob
import swift.common.utils
from swift.common.request_helpers import is_sys_meta

try:
    import swift.common.splice
    HAS_NEW_SPLICE = True
except ImportError:
    HAS_NEW_SPLICE = False

from scality_sproxyd_client.exceptions import SproxydHTTPException
import swift_scality_backend.http_utils
import swift_scality_backend.splice_utils
from swift_scality_backend import utils

# These are system-set metadata keys that cannot be changed with a POST.
# They should be lowercase.
RESERVED_DATAFILE_META = {'content-length', 'deleted', 'etag'}
DATAFILE_SYSTEM_META = {'x-static-large-object'}


class DiskFileWriter(object):
    """A simple sproxyd pass-through

    Encapsulation of the write context for servicing PUT REST API
    requests. Serves as the context manager object for DiskFile's create()
    method.

    :param client_collection: Client collection to use to perform operations
    :type client_collection: `swift_scality_backend.http_utils.ClientCollection`
    :param name: standard object name
    :type name: `str`
    :param logger: Logger to use within the `DiskFileWriter`
    :type logger: `logging.Logger`
    """
    def __init__(self, client_collection, name, logger):
        self._client_collection = client_collection
        self._name = name
        self._logger = logger
        self._upload_size = 0
        self._md5sum = hashlib.md5()

        headers = {
            'transfer-encoding': 'chunked'
        }
        self.logger.debug("DiskFileWriter for %r initialized", self._name)

        client = self._client_collection.get_write_client()
        self._conn, self._release_conn = client.get_http_conn_for_put(
            self._name, headers)

    def __repr__(self):
        ret = 'DiskFileWriter(client_collection=%r, name=%r, logger=%r)'
        return ret % (self._client_collection, self._name, self._logger)

    logger = property(operator.attrgetter('_logger'))

    def write(self, chunk):
        """Write a chunk of data.

        :param chunk: the chunk of data to write as a string object
        """
        self._conn.send('%x\r\n%s\r\n' % (len(chunk), chunk))
        self._upload_size += len(chunk)
        self._md5sum.update(chunk)
        return self._upload_size

    @utils.trace
    def put(self, metadata):
        """Finalize writing the object.

        :param metadata: dictionary of metadata to be associated with the
                         object
        """
        self._conn.send('0\r\n\r\n')
        try:
            resp = self._conn.getresponse()
            msg = resp.read()
            if resp.status != 200:
                raise SproxydHTTPException("putting: %s / %s" % (
                    str(resp.status), str(msg)))
        except Exception:
            conn, self._conn = self._conn, None
            try:
                conn.close()
            except Exception:
                self.logger.exception('Failure while closing connection')
            raise

        self._release_conn()

        metadata_to_put = copy.copy(metadata)
        metadata_to_put['name'] = self._name
        metadata_to_put['ETag'] = self._md5sum.hexdigest()

        metadata_to_put.update({
            'df': metadata,
            'mf': {},
        })

        self.logger.debug("Data successfully written for object : %r", self._name)

        self._client_collection.get_write_client().put_meta(
            self._name, metadata_to_put)

    def commit(self, timestamp):
        """
        Perform any operations necessary to mark the object as durable.

        :param timestamp: object put timestamp, an instance of
                          :class:`~swift.common.utils.Timestamp`
        """
        pass


class DiskFileReader(object):
    """A simple sproxyd pass-through

    Encapsulation of the read context for servicing GET REST API
    requests. Serves as the context manager object for DiskFile's reader()
    method.

    :param filesystem: internal file system object to use
    :param name: object name
    :param use_splice: if true, use zero-copy splice() to send data
    """
    def __init__(self, client_collection, name, use_splice, logger):
        self._client_collection = client_collection
        self._name = name
        self._use_splice = use_splice
        self._logger = logger

    def __repr__(self):
        ret = (
            'DiskFileReader(client_collection=%r, name=%r, use_splice=%r, '
            'logger=%r)')
        return ret % (self._client_collection, self._name, self._use_splice,
                      self.logger)

    logger = property(operator.attrgetter('_logger'))

    @utils.trace
    def __iter__(self):
        headers, data = self._client_collection.try_read(
            lambda client: client.get_object(self._name))
        return data

    @utils.trace
    def can_zero_copy_send(self):
        return self._use_splice

    @utils.trace
    def zero_copy_send(self, wsockfd):
        client = self._client_collection.get_read_client()

        object_url = urlparse.urlparse(client.get_url_for_object(self._name))
        conn = None

        with swift.common.exceptions.ConnectionTimeout(client.conn_timeout):
            conn = swift_scality_backend.http_utils.SomewhatBufferedHTTPConnection(
                object_url.netloc)

            try:
                conn.putrequest('GET', object_url.path, skip_host=False)
                conn.endheaders()
            except:  # noqa
                conn.close()
                raise

        with conn:
            resp = conn.getresponse()

            if resp.status != httplib.OK:
                raise SproxydHTTPException(
                    'Unexpected response code: %s' % resp.status,
                    url=object_url.geturl(),
                    http_status=resp.status, http_reason=resp.reason)

            if resp.chunked:
                raise SproxydHTTPException(
                    'Chunked response not supported',
                    url=object_url.geturl(),
                    http_status=resp.status, http_reason=resp.reason)

            buff = resp.fp.get_buffered()
            buff_len = len(buff)

            while buff:
                written = eventlet.green.os.write(wsockfd, buff)
                buff = buff[written:]

            to_splice = resp.length - buff_len if resp.length is not None else None

            swift_scality_backend.splice_utils.splice_socket_to_socket(
                resp.fileno(), wsockfd, length=to_splice)

    @utils.trace
    def app_iter_range(self, start, stop):
        """
        Iterate over a range.

        :param start: First byte to read from (inclusive)
        :type start: int
        :param stop: Last byte to read (exclusive)
        :type stop: int
        """
        # HTTP range is inclusive on both ends.
        headers = {
            'range': 'bytes=%d-%d' % (start, stop - 1)
        }

        headers, data = self._client_collection.try_read(
            lambda client: client.get_object(self._name, headers))
        return data

    @utils.trace
    def app_iter_ranges(self, ranges, content_type, boundary, size):
        """iterate over multiple ranges."""
        if not ranges:
            yield ''
        else:
            for chunk in swift.common.swob.multi_range_iterator(
                    ranges, content_type, boundary, size, self.app_iter_range):
                yield chunk


class DiskFile(object):
    """A simple sproxyd pass-through

    :param client_collection: Client collection to use to perform operations
    :type client_collection: `swift_scality_backend.http_utils.ClientCollection`
    :param account: account name for the object
    :type account: `str`
    :param container: container name for the object
    :type container: `str`
    :param obj: object name for the object
    :type obj: `str`
    :param use_splice: if true, use zero-copy splice() to send data
    :type use_splice: `bool`
    """

    def __init__(self, client_collection, account, container, obj, use_splice,
                 logger):
        # We hash the account, container and object name so that no 'special'
        # character will get in our way.
        sha1 = hashlib.sha1()
        for part in [account, container, obj]:
            sha1.update(part)
        self._name = sha1.hexdigest()
        self._metadata = None
        self._client_collection = client_collection
        self._logger = logger

        self._account = account
        self._container = container
        self._obj = obj
        self._use_splice = use_splice

    logger = property(operator.attrgetter('_logger'))
    client_collection = property(operator.attrgetter('_client_collection'))

    def __repr__(self):
        ret = ('DiskFile(client_collection=%r, account=%r, container=%r, obj=%r, '
               'use_splice=%r, logger=%r)')
        return ret % (self._client_collection, self._account, self._container,
                      self._obj, self._use_splice, self._logger)

    @staticmethod
    def merge_df_mf_metadata(df_md_source, mf_md_source):
        """Merge the datafile metadata and metafile metadata dictionaries together.

        Datafile metadata refers to metadata originally included when the object
        was first PUT, and does not include metadata set by any subsequent POST.

        Metafile metadata refers to metadata written by a POST, and does not
        include any persistent metadata that was set by the original PUT.
        """
        md_dest = {}

        if not mf_md_source:
            md_dest.update(df_md_source)
        else:
            sys_metadata = {
                key: val for key, val in df_md_source.items()
                if key.lower() in (RESERVED_DATAFILE_META | DATAFILE_SYSTEM_META) or
                is_sys_meta('object', key)
            }
            md_dest.update(mf_md_source)
            md_dest.update(sys_metadata)

        return md_dest

    @utils.trace
    def open(self, current_time=None):
        """Open the file and read the metadata.

        :param current_time: Unix time used in checking expiration. If not
             present, the current time will be used.

        This method must populate the _metadata attribute.

        :raise DiskFileDeleted: if it does not exist
        """
        try:
            metadata = self.client_collection.try_read(
                lambda client: client.get_meta(self._name))
        except EOFError:
            self.logger.error('ERROR in DiskFile.open(): metadata not found' +
                              ' on Scality RING for key %s' % self._name)
            metadata = None

        if metadata is None:
            raise swift.common.exceptions.DiskFileDeleted()

        # 'df' subdictionary refers to DataFile metadata, 'mf' to MetaFile metadata
        if 'df' in metadata and 'mf' in metadata:
            self._metadata = DiskFile.merge_df_mf_metadata(metadata['df'], metadata['mf'])
            self._metadata['df'] = metadata['df']
            self._metadata['mf'] = metadata['mf']
        else:
            self._metadata = metadata
            self._metadata['df'] = copy.copy(metadata)
            self._metadata['mf'] = {}

        try:
            x_delete_at = int(self._metadata['X-Delete-At'])
        except KeyError:
            pass
        else:
            if current_time is None:
                current_time = time.time()
            if x_delete_at <= current_time:
                raise swift.common.exceptions.DiskFileExpired(
                    metadata=self._metadata)

        return self

    @utils.trace
    def __enter__(self):
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        return self

    @utils.trace
    def __exit__(self, t, v, tb):
        pass

    @utils.trace
    def get_metadata(self):
        """Provide the metadata for an object as a dictionary.

        :returns: object's metadata dictionary
        """
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        md_to_return = copy.deepcopy(self._metadata)
        md_to_return.pop('df')
        md_to_return.pop('mf')
        return md_to_return

    @utils.trace
    def read_metadata(self, current_time=None):
        """Return the metadata for an object without requiring the caller
        to open the object first.

        :param current_time: Unix time used in checking expiration. If not
             present, the current time will be used.
        :returns: metadata dictionary for an object
        """
        with self.open(current_time=current_time):
            return self.get_metadata()

    @utils.trace
    def reader(self, keep_cache=False):
        """Return a swift.common.swob.Response class compatible "app_iter" object.

        :param keep_cache: ignored. Kept for compatibility with the native
                          `DiskFile` class in Swift
        """
        dr = DiskFileReader(self.client_collection, self._name,
                            use_splice=self._use_splice,
                            logger=self.logger)
        return dr

    @utils.trace
    @contextlib.contextmanager
    def create(self, size=None):
        """Context manager to create a file.

        :param size: ignored. Kept for compatibility with the native
                     `DiskFile` class in Swift. This `create` method is
                     called externally only by the `ObjectController`
        """
        yield DiskFileWriter(self.client_collection, self._name, self.logger)

    @utils.trace
    def write_metadata(self, md_to_add):
        """Write a block of metadata to an object."""

        if not self._metadata:
            md_to_write = copy.deepcopy(md_to_add)
            md_to_write.update({
                'df': {
                    'name': self._name,
                },
                'mf': md_to_add
            })
            self.client_collection.get_write_client().put_meta(self._name, md_to_write)
            return

        # Keep the most recent Content-Type from either datafile or metafile metadata
        if 'Content-Type' not in md_to_add and self._metadata is not None and \
           'Content-Type' in self._metadata['mf'] and \
           self._metadata['mf']['Content-Type'] != self._metadata['df']['Content-Type'] and \
           self._metadata['mf']['Content-Type-Timestamp'] > \
           self._metadata['df']['X-Timestamp']:
            md_to_add['Content-Type'] = self._metadata['mf']['Content-Type']
            md_to_add['Content-Type-Timestamp'] = self._metadata['mf']['Content-Type-Timestamp']

        df_md_dict = copy.deepcopy(self._metadata['df'])
        md_to_write = DiskFile.merge_df_mf_metadata(df_md_dict, md_to_add)
        md_to_write.update({
            'df': df_md_dict,
            'mf': md_to_add,
        })

        self.client_collection.get_write_client().put_meta(self._name, md_to_write)

    @utils.trace
    def delete(self, timestamp):
        """Perform a delete for the given object in the given container under
        the given account.

        :param timestamp: ignored. Kept for compatibility with the native
                          `DiskFile` class in Swift. This `delete` method is
                          called externally only by the `ObjectController`
        """
        self.client_collection.get_write_client().del_object(self._name)

    def get_metafile_metadata(self):
        """Provide the metafile metadata for a previously opened object as a dictionary.

        This is metadata that was written by a POST,
        and does not include any persistent metadata that was set by the original PUT.
        """
        if self._metadata is None:
            self._metadata = self.client_collection.try_read(
                lambda client: client.get_meta(self._name))
        return self._metadata['mf']

    def get_datafile_metadata(self):
        """Provide the datafile metadata for a previously opened object as a dictionary.

        This is metadata that was included when the object was first PUT,
        and does not include metadata set by any subsequent POST.
        """
        if self._metadata is None:
            self._metadata = self.client_collection.try_read(
                lambda client: client.get_meta(self._name))
        return self._metadata['df']

    @property
    def content_length(self):
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        return self._metadata.get('Content-Length')

    @property
    def timestamp(self):
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        t = self._metadata.get('X-Timestamp')
        return swift.common.utils.Timestamp(t)

    @property
    def data_timestamp(self):
        """Provides the datafile timestamp
        """
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        t = self._metadata['df'].get('X-Timestamp')
        return swift.common.utils.Timestamp(t)

    @property
    def durable_timestamp(self):
        """Related to erasure coding storage policy type, not to replication
        policy type.

        Provides the timestamp of the newest data file found in the object
        directory, i.e. the one which call to open() populated the self._metadata
        attribute.
        """
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        t = self._metadata['df'].get('X-Timestamp')
        return swift.common.utils.Timestamp(t)

    @property
    def fragments(self):
        return None

    @property
    def content_type(self):
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        return self._metadata.get('Content-Type')

    @property
    def content_type_timestamp(self):
        """Provides the content-type timestamp if the Content-Type metadata
        attribute was ever modified, or the datafile metadata timestamp if not.
        """
        if self._metadata is None:
            raise swift.common.exceptions.DiskFileNotOpen()
        t = self._metadata.get('Content-Type-Timestamp',
                               self._metadata['df'].get('X-Timestamp'))
        return swift.common.utils.Timestamp(t)


class DiskFileManager(object):
    """
    Management class for devices, providing common place for shared parameters
    and methods not provided by the DiskFile class (which primarily services
    the object server REST API layer).

    The `get_diskfile()` method is how this implementation creates a `DiskFile`
    object.
    """

    def __init__(self, conf, logger):
        """
        :param conf: caller provided configuration object
        :param logger: caller provided logger
        """
        self.logger = logger

        self.use_splice = False

        conf_wants_splice = swift.common.utils.config_true_value(
            conf.get('splice', 'no'))

        if HAS_NEW_SPLICE:
            system_has_splice = swift.common.splice.splice.available
        else:
            try:
                system_has_splice = swift.common.utils.system_has_splice()
            except AttributeError:  # Old Swift versions
                system_has_splice = False

        if conf_wants_splice and not system_has_splice:
            self.logger.warn(
                "Use of splice() requested (config says \"splice = %s\"), "
                "but the system does not support it. "
                "splice() will not be used." % conf.get('splice'))

        if conf_wants_splice and system_has_splice:
            self.use_splice = True

    def get_diskfile(self, client_collection, account, container, obj):
        return DiskFile(client_collection, account, container, obj,
                        use_splice=self.use_splice, logger=self.logger)

    def pickle_async_update(self, *args, **kwargs):
        pass



# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import logging
import sys

DEFAULT_LOGGER = logging.getLogger(__name__)


def trace(f):
    '''Trace calls to a decorated function

    Using this decorator on a function will cause its execution to be logged at
    `DEBUG` level, including messages when the function is called (with a call
    identifier and the arguments), as well as when the function returns (also
    including the call identifier, and a representation of the return value or
    exception if applicable).
    '''

    # 'State' hack
    tid = [0]

    name = f.func_name

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        maybe_self = None

        if len(args) > 0:
            maybe_self = args[0]
        else:
            maybe_self = kwargs.get('self', None)

        assert maybe_self is not None

        logger = getattr(maybe_self, 'logger', DEFAULT_LOGGER)

        # Fast-path, don't do any of the 'expensive' things below if the debug
        # log level isn't enabled anyway.
        # The conditional method lookup is for compatibility with the
        # `LoggerAdapter` implementation in Python 2.6, used by Swift, which
        # lacks the `isEnabledFor` method (added in Python 2.7). We assume the
        # log level is enabled in that case.
        if not getattr(logger, 'isEnabledFor', lambda _: True)(logging.DEBUG):
            return f(*args, **kwargs)

        # Get & bump call identifier, assume non-preemptive threading
        ctid, tid[0] = tid[0], tid[0] + 1

        if sys.version_info >= (2, 7):
            all_args = inspect.getcallargs(f, *args, **kwargs)
            logger.debug('==> %s (%d): call %r', name, ctid, all_args)
        else:
            logger.debug('==> %s (%d): call', name, ctid)

        result = None

        try:
            result = f(*args, **kwargs)
        except BaseException as exc:
            logger.debug('<== %s (%d): exception %r', name, ctid, exc)
            raise
        else:
            logger.debug('<== %s (%d): return %r', name, ctid, result)
            return result

    return wrapped


def split_list(val):
    '''Split a comma-separated string into a list of strings.'''

    return (s2 for s2 in
            (s1.strip() for s1 in val.split(','))
            if s2)



# Copyright (c) 2014, Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.server`'''

import errno
import inspect
import os
import StringIO
import unittest
import urlparse

import mock

import swift.obj.server

import swift_scality_backend.diskfile
import swift_scality_backend.policy_configuration
import swift_scality_backend.server

from . import utils


def test_api_compatible():
    '''
    Test whether `swift_scality_backend.server.app_factory`'s result is
    API-compatible with `swift.obj.server.app_factory`'s result
    '''
    whitelist = [
        'replication_failure_threshold',
        'replication_semaphore',
        'replication_failure_ratio',
    ]

    swift_server = swift.obj.server.app_factory({})
    conf = {'sproxyd_host': 'host1:81'}
    scality_server = swift_scality_backend.server.app_factory(conf)

    def assert_compatible(name, spec1, spec2):
        '''Assert argspecs are compatible'''

        if spec1.varargs:
            assert spec2.varargs, 'No varargs: %r' % name

        if spec1.keywords:
            assert spec2.keywords, 'No kwargs: %r' % name

        nargs1 = len(spec1.args)
        ndefs1 = len(spec1.defaults or [])
        nargs2 = len(spec2.args)
        ndefs2 = len(spec2.defaults or [])

        assert nargs2 >= nargs1, 'Less args: %r' % name
        assert nargs2 - ndefs2 <= nargs1 - ndefs1, \
            'Incompatible number of non-default args: %r' % name

        # The `policy` arg used to be named  `policy_index` or `policy_idx`
        # in Swift 2.1 and 2.2. It changed in Swift 2.3.
        # We rename the arg here to be keep compatibility and have the same
        # method signature excepted for the name of the `policy` argument.
        spec1_args, spec2_args = spec1.args, spec2.args
        replace = {'policy_idx': 'policy', 'policy_index': 'policy'}
        if name in ['get_diskfile', 'async_update']:
            spec1_args = [replace.get(arg, arg) for arg in spec1.args]
            spec2_args = [replace.get(arg, arg) for arg in spec2.args]

        assert spec1_args == spec2_args[:nargs1], \
            'Incompatible arg names: %r' % name

    def check_api_compatible(name):
        '''Check whether the API of a given method name is compatible'''

        assert hasattr(scality_server, name), 'Missing attribute: %r' % name

        swift_attr = getattr(swift_server, name)
        scality_attr = getattr(scality_server, name)

        if callable(swift_attr):
            assert callable(scality_attr), 'Not callable: %r' % name

            swift_spec = inspect.getargspec(swift_attr)
            scality_spec = inspect.getargspec(scality_attr)

            assert_compatible(name, swift_spec, scality_spec)

    for name in dir(swift_server):
        if name in whitelist or name[0] == '_':
            continue

        yield check_api_compatible, name


def test_setup_with_sproxyd_endpoints():
    endpoint1 = 'http://h1:81/p'
    endpoint2 = 'http://h2:81/p'
    conf = {'sproxyd_endpoints': ' , '.join([endpoint1, endpoint2])}
    obj_serv = swift_scality_backend.server.app_factory(conf)

    assert frozenset([
        urlparse.urlparse(endpoint1),
        urlparse.urlparse(endpoint2),
    ]) == obj_serv._get_client_for_policy(0).read_clients[0]._endpoints


def test_setup_with_custom_timeout():
    conf = {'sproxyd_host': 'host1:81', 'sproxyd_proxy_timeout': "10.0",
            'sproxyd_conn_timeout': "4.1"}
    obj_serv = swift_scality_backend.server.app_factory(conf)

    assert obj_serv._get_client_for_policy(0).read_clients[0].read_timeout == 10.0
    assert obj_serv._get_client_for_policy(0).read_clients[0].conn_timeout == 4.1


def test_get_diskfile():
    conf = {'sproxyd_host': 'host1:81'}
    scality_server = swift_scality_backend.server.app_factory(conf)
    diskfile = scality_server.get_diskfile('dev', 'partition', 'a', 'c', 'o')

    assert isinstance(diskfile, swift_scality_backend.diskfile.DiskFile)
    assert diskfile._account == 'a'
    assert diskfile._container == 'c'
    assert diskfile._obj == 'o'


_mock_policy_config_file = mock.mock_open()

_real_open = open


def _open(name, arg):
    if name == os.devnull:
        return _real_open(name, arg)
    else:
        raise IOError(errno.ENOENT, 'Mock')


_mock_policy_config_file.side_effect = _open

del _open


class FakeFile(StringIO.StringIO):
    def __enter__(self):
        return self

    def __exit__(self, *_):
        pass


@mock.patch('eventlet.spawn', mock.Mock())
@mock.patch('__builtin__.open', _mock_policy_config_file)
class TestStoragePolicySupport(unittest.TestCase):
    @staticmethod
    def _app_factory(**kwargs):
        real_kwargs = {
            'sproxyd_host': '127.0.0.1:81',
        }
        real_kwargs.update(kwargs)

        return swift_scality_backend.server.app_factory(real_kwargs)

    def test_no_policy(self):
        with mock.patch(
                'swift_scality_backend.server.ObjectController._get_client_for_policy') as mock_gc:
            self._app_factory().get_diskfile('dev', 'partition', 'a', 'c', 'o')

            mock_gc.assert_called_with(0)

    def test_policy_0(self):
        with mock.patch(
                'swift_scality_backend.server.ObjectController._get_client_for_policy') as mock_gc:
            self._app_factory().get_diskfile('dev', 'partition', 'a', 'c', 'o', policy_idx=0)

            mock_gc.assert_called_with(0)

    def test_policy_1(self):
        with mock.patch(
                'swift_scality_backend.server.ObjectController._get_client_for_policy') as mock_gc:
            self._app_factory().get_diskfile('dev', 'partition', 'a', 'c', 'o', policy_idx=1)

            mock_gc.assert_called_with(1)

    def test_policy_file_read_error(self):
        mock_open = mock.mock_open()
        mock_open.side_effect = IOError(errno.ENFILE, os.strerror(errno.ENFILE))

        with mock.patch('__builtin__.open', mock_open):
            self.assertRaises(
                IOError,
                self._app_factory)

    def test_policy_request_without_configuration(self):
        self.assertRaises(
            RuntimeError,
            self._app_factory().get_diskfile, 'dev', 'partition', 'a', 'c', 'o', policy_idx=1)

    def test_broken_policy_configuration(self):
        policy = '\n'.join(line.strip() for line in '''
            [storage-policy:1]
            write = no-such-ring
            '''.splitlines())

        mock_open = mock.mock_open()
        mock_open.return_value = FakeFile(policy)

        with mock.patch('__builtin__.open', mock_open):
            utils.assertRaisesRegexp(
                swift_scality_backend.policy_configuration.ConfigurationError,
                'Unknown \'write\' ring \'no-such-ring\' in policy 1',
                self._app_factory)

    def test_simple_policy_configuration(self):
        policy = '\n'.join(line.strip() for line in '''
            [ring:paris]
            location = paris
            sproxyd_endpoints = http://localhost:8080/chord, http://otherhost:8080/chord

            [storage-policy:1]
            read =
            write = paris
            '''.splitlines())

        mock_open = mock.mock_open()
        mock_open.return_value = FakeFile(policy)

        with mock.patch('__builtin__.open', mock_open):
            df = self._app_factory().get_diskfile(
                'dev', 'partition', 'a', 'c', 'o', policy_idx=1)

            self.assertEqual(frozenset([
                urlparse.urlparse('http://localhost:8080/chord'),
                urlparse.urlparse('http://otherhost:8080/chord'),
            ]), df._client_collection.read_clients[0]._alive)



# Copyright (c) 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.policy_configuration`.'''

import unittest
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import utils

from swift_scality_backend.policy_configuration import Configuration
from swift_scality_backend.policy_configuration import ConfigurationError
from swift_scality_backend.policy_configuration import Endpoint
from swift_scality_backend.policy_configuration import Location
from swift_scality_backend.policy_configuration import Ring
from swift_scality_backend.policy_configuration import StoragePolicy


class TestEndpoint(unittest.TestCase):
    def test_hash_equal(self):
        url = 'http://localhost:81/path'

        self.assertEqual(
            hash(Endpoint(url)),
            hash(Endpoint(url)))

    def test_hash_not_equal(self):
        # If this fails, the subsequent test would fail as well, unexpectedly
        self.assertNotEqual(
            hash('http://localhost/path'),
            hash('http://otherhost/path'))

        self.assertNotEqual(
            hash(Endpoint('http://localhost/path')),
            hash(Endpoint('http://otherhost/path')))

    def test_eq_not_implemented(self):
        # Note: this can't be `assertNotEqual`, we want to test `__eq__`
        self.assertFalse(Endpoint('http://localhost') == 1)

    def test_ne_not_implemented(self):
        self.assertNotEqual(Endpoint('http://localhost'), 1)

    def test_reject_params(self):
        utils.assertRaisesRegexp(
            ValueError,
            r'Endpoint URL can\'t have params',
            Endpoint, 'http://localhost/;param')

    def test_reject_query(self):
        utils.assertRaisesRegexp(
            ValueError,
            r'Endpoint URL can\'t have query values',
            Endpoint, 'http://localhost/?query')

    def test_reject_fragment(self):
        utils.assertRaisesRegexp(
            ValueError,
            r'Endpoint URL can\'t have a fragment',
            Endpoint, 'http://localhost/#fragment')


class TestLocation(unittest.TestCase):
    def test_eq_not_implemented(self):
        self.assertFalse(Location('paris') == 1)

    def test_ne_not_implemented(self):
        self.assertNotEqual(Location('paris'), 1)

    def test_str(self):
        self.assertEqual('paris', str(Location('paris')))

    def test_repr(self):
        self.assertEqual('Location(name=\'paris\')', repr(Location('paris')))

    def test_hash_equal(self):
        self.assertEqual(hash(Location('paris')), hash(Location('paris')))

    def test_hash_not_equal(self):
        # Require this first... If this fails, our test would fail as well which
        # is not the intention
        self.assertNotEqual(hash('paris'), hash('london'))
        self.assertNotEqual(hash(Location('paris')), hash(Location('london')))


class TestRing(unittest.TestCase):
    def test_eq(self):
        name = 'paris-arc6+3'
        location = 'paris'
        url1 = 'http://localhost'
        url2 = 'http://otherhost'

        r1 = Ring(name, location, [url1, url2])
        r2 = Ring(name, location, [url2, url1])

        self.assertEqual(r1, r2)

    def test_eq_not_implemented(self):
        self.assertFalse(Ring('paris-arc6+3', 'paris', []) == 1)

    def test_ne(self):
        self.assertNotEqual(
            Ring('paris-arc6+3', 'paris', []),
            Ring('london-arc6+3', 'london', []))

    def test_ne_not_implemented(self):
        self.assertNotEqual(Ring('paris-arc6+3', 'paris', []), 1)

    def test_hash_equal(self):
        name = 'paris-arc6+3'
        location = 'paris'
        url1 = 'http://localhost'
        url2 = 'http://otherhost'

        r1 = Ring(name, location, [url1, url2])
        r2 = Ring(name, location, [url2, url1])

        self.assertEqual(hash(r1), hash(r2))

    def test_hash_not_equal(self):
        self.assertNotEqual(hash('paris-arc6+3'), hash('paris-chord3'))

        self.assertNotEqual(
            hash(Ring('paris-arc6+3', 'paris', [])),
            hash(Ring('paris-chord3', 'paris', [])))


class TestStoragePolicy(unittest.TestCase):
    def test_eq(self):
        index = 1
        read_set = [Ring('paris-arc6+3', 'paris', ['http://paris'])]
        write_set = [Ring('london-arc6+3', 'london', ['http://london'])]

        sp1 = StoragePolicy(index, read_set, write_set)
        sp2 = StoragePolicy(index, read_set, write_set)

        self.assertEqual(sp1, sp2)

    def test_eq_not_implemented(self):
        self.assertFalse(StoragePolicy(1, [], []) == 1)

    def test_ne(self):
        self.assertNotEqual(
            StoragePolicy(1, [], []),
            StoragePolicy(2, [], []))

        self.assertNotEqual(
            StoragePolicy(1, [Ring('paris-arc6+3', 'paris', [])], []),
            StoragePolicy(1, [Ring('london-arc6+3', 'london', [])], []))

        self.assertNotEqual(
            StoragePolicy(1, [], [Ring('paris-arc6+3', 'paris', [])]),
            StoragePolicy(1, [], []))

    def test_ne_not_implemented(self):
        self.assertNotEqual(StoragePolicy(1, [], []), 1)

    def test_hash_equal(self):
        self.assertEqual(
            hash(
                StoragePolicy(
                    1,
                    [
                        Ring('paris-arc6+3', 'paris',
                             ['http://paris1', 'http://paris2']),
                        Ring('london-arc6+3', 'london',
                             ['http://london1', 'http://london2'])],
                    [
                        Ring('paris-chord3', 'paris', ['http://paris3'])])),
            hash(
                StoragePolicy(
                    1,
                    [
                        Ring('london-arc6+3', 'london',
                             ['http://london2', 'http://london1']),
                        Ring('paris-arc6+3', 'paris',
                             ['http://paris1', 'http://paris2'])],
                    [
                        Ring('paris-chord3', 'paris', ['http://paris3'])])))

    def test_hash_not_equal(self):
        self.assertNotEqual(
            hash(StoragePolicy(1, [], [])),
            hash(StoragePolicy(2, [], [])))

    LOOKUP_TEST_POLICY = StoragePolicy(
        index=1,
        read_set=[
            Ring(
                name='sfo-arc6+3',
                location=Location(name='sfo'),
                endpoints=[
                    Endpoint(url='http://sfo1.int/arc6+3')])],
        write_set=[
            Ring(
                name='paris-arc6+3',
                location=Location(name='paris'),
                endpoints=[
                    Endpoint(url='http://paris1.int/arc6+3'),
                    Endpoint(url='http://paris2.int/arc6+3')])])

    def test_lookup_method_invalid(self):
        self.assertRaises(ValueError, StoragePolicy(1, [], []).lookup, 'test')

    def test_lookup_method_read_without_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.READ),
            [frozenset([
                Endpoint('http://sfo1.int/arc6+3'),
                Endpoint('http://paris1.int/arc6+3'),
                Endpoint('http://paris2.int/arc6+3')])])

    def test_lookup_method_write_without_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.WRITE),
            [frozenset([
                Endpoint('http://paris1.int/arc6+3'),
                Endpoint('http://paris2.int/arc6+3')])])

    def test_lookup_method_read_with_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.READ, ['paris']),
            [
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3')]),
                frozenset([
                    Endpoint('http://sfo1.int/arc6+3')])])

        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.READ, [Location('sfo')]),
            [
                frozenset([
                    Endpoint('http://sfo1.int/arc6+3')]),
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3')])])

    def test_lookup_method_write_with_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.WRITE, ['paris']),
            [
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3')])])

        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.WRITE, ['sfo']),
            [
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3')])])

    def test_lookup_method_read_with_unknown_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.READ, ['nyc']),
            [
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3'),
                    Endpoint('http://sfo1.int/arc6+3')])])

    def test_lookup_method_write_with_unknown_location(self):
        self.assertEqual(
            self.LOOKUP_TEST_POLICY.lookup(StoragePolicy.WRITE, ['nyc']),
            [
                frozenset([
                    Endpoint('http://paris1.int/arc6+3'),
                    Endpoint('http://paris2.int/arc6+3')])])


class TestConfiguration(unittest.TestCase):
    TEST_CONFIGURATION = '\n'.join(line.lstrip() for line in '''
        [ring:paris-rep3]
        location = paris
        sproxyd_endpoints = http://paris1.int/rep3, http://paris2.int/rep3

        [ring:paris-arc6+3]
        location = paris
        sproxyd_endpoints = http://paris1.int/arc6+3, http://paris2.int/arc6+3

        [ring:sfo-arc6+3]
        location = sfo
        sproxyd_endpoints = http://sfo1.int/arc6+3

        [ring:nyc-arc6+3]
        location = nyc
        sproxyd_endpoints = http://nyc1.int/arc6+3

        [storage-policy:1]
        read = sfo-arc6+3
        write = paris-arc6+3

        [storage-policy:2]
        read = nyc-arc6+3
        write = paris-arc6+3

        [storage-policy:3]
        read =
        write = paris-rep3
        '''.splitlines())

    def test_duplicate_index(self):
        p1 = StoragePolicy(1, [], [])
        p2 = StoragePolicy(2, [], [])
        p3 = StoragePolicy(1, [], [])

        self.assertRaises(ValueError, Configuration, [p1, p2, p3])

    def test_eq(self):
        self.assertEqual(
            Configuration([StoragePolicy(1, [], [])]),
            Configuration([StoragePolicy(1, [], [])]))

    def test_eq_not_implemented(self):
        self.assertFalse(
            Configuration([]) == 1)

    def test_ne(self):
        self.assertNotEqual(
            Configuration([StoragePolicy(1, [], [])]),
            Configuration([]))

    def test_ne_not_implemented(self):
        self.assertNotEqual(Configuration([]), 1)

    def test_hash_equal(self):
        p = StoragePolicy(1, [], [])

        self.assertEqual(
            hash(Configuration([p])),
            hash(Configuration([p])))

    def test_from_stream(self):
        conf = Configuration.from_stream(StringIO(self.TEST_CONFIGURATION))

        p1 = conf.get_policy(1)
        self.assertEqual(p1.index, 1)
        self.assertEqual(
            list(p1.read_set),
            [Ring('sfo-arc6+3', 'sfo', ['http://sfo1.int/arc6+3'])])
        self.assertEqual(
            list(p1.write_set),
            [Ring(
                'paris-arc6+3', 'paris',
                ['http://paris1.int/arc6+3', 'http://paris2.int/arc6+3'])])

        p2 = conf.get_policy(2)
        self.assertEqual(
            p2,
            StoragePolicy(
                2,
                [Ring('nyc-arc6+3', 'nyc', ['http://nyc1.int/arc6+3'])],
                [Ring(
                    'paris-arc6+3', 'paris',
                    ['http://paris1.int/arc6+3', 'http://paris2.int/arc6+3'])]))

        p3 = conf.get_policy(3)
        self.assertEqual(
            p3,
            StoragePolicy(
                3,
                [],
                [Ring(
                    'paris-rep3', 'paris',
                    ['http://paris1.int/rep3', 'http://paris2.int/rep3'])]))

        self.assertRaises(ValueError, conf.get_policy, 4)
        self.assertRaises(ValueError, conf.get_policy, 'test')

    def test_to_stream(self):
        conf = Configuration.from_stream(StringIO(self.TEST_CONFIGURATION))

        out = StringIO()
        conf.to_stream(out)

        conf2 = Configuration.from_stream(StringIO(out.getvalue()))

        self.assertEqual(conf, conf2)

    @staticmethod
    def _format_config(conf):
        return StringIO(
            '\n'.join(
                line.lstrip() for line in conf.splitlines()))

    def test_missing_ring_name(self):
        conf = self._format_config('''
            [ring:]
            location = test
            sproxyd_endpoints = http://localhost
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Invalid section name \'ring:\'',
            Configuration.from_stream, conf)

    def test_missing_ring_location(self):
        conf = self._format_config('''
            [ring:test]
            sproxyd_endpoints = http://localhost
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            '''Section 'ring:test' lacks a 'location' setting''',
            Configuration.from_stream, conf)

    def test_empty_ring_location(self):
        conf = self._format_config('''
            [ring:test]
            location =
            sproxyd_endpoints = http://localhost
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Invalid \'location\' setting in \'ring:test\'',
            Configuration.from_stream, conf)

    def test_empty_ring_sproxyd_endpoints(self):
        conf = self._format_config('''
            [ring:test]
            location = test
            sproxyd_endpoints =
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Invalid \'sproxyd_endpoints\' setting in \'ring:test\'',
            Configuration.from_stream, conf)

    def test_missing_ring_sproxyd_endpoints(self):
        conf = self._format_config('''
            [ring:test]
            location = paris
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            '''Section 'ring:test' lacks a 'sproxyd_endpoints' setting''',
            Configuration.from_stream, conf)

    def test_invalid_ring_sproxyd_endpoint(self):
        conf = self._format_config('''
            [ring:test]
            location = paris
            sproxyd_endpoints = http://localhost, http://otherhost/?a=b
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Error parsing endpoint \'http://otherhost/\?a=b\' in '
            '\'ring:test\': Endpoint URL can\'t have query values',
            Configuration.from_stream, conf)

    def test_non_int_storage_policy_index(self):
        conf = self._format_config('''
            [storage-policy:test]
            read =
            write = test
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Invalid policy index: \'test\'',
            Configuration.from_stream, conf)

    def test_empty_storage_policy_index(self):
        conf = self._format_config('''
            [storage-policy:]
            read =
            write = test
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Invalid section name \'storage-policy:\'',
            Configuration.from_stream, conf)

    def test_unknown_write_ring_in_storage_policy(self):
        conf = self._format_config('''
            [ring:test1]
            location = paris
            sproxyd_endpoints = http://localhost

            [storage-policy:1]
            read =
            write = test2
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Unknown \'write\' ring \'test2\' in policy 1',
            Configuration.from_stream, conf)

    def test_unknown_read_ring_in_storage_policy(self):
        conf = self._format_config('''
            [ring:test1]
            location = paris
            sproxyd_endpoints = http://localhost

            [storage-policy:1]
            read = test2
            write = test1
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Unknown \'read\' ring \'test2\' in policy 1',
            Configuration.from_stream, conf)

    def test_multiple_write_rings_in_storage_policy(self):
        conf = self._format_config('''
            [ring:test1]
            location = paris
            sproxyd_endpoints = http://paris

            [ring:test2]
            location = london
            sproxyd_endpoints = http://london

            [storage-policy:1]
            read =
            write = test1, test2
            ''')

        utils.assertRaisesRegexp(
            ConfigurationError,
            'Multiple \'write\' rings defined in \'storage-policy:1\'',
            Configuration.from_stream, conf)

    def test_missing_read_in_storage_policy(self):
        conf = self._format_config('''
            [ring:test1]
            location = paris
            sproxyd_endpoints = http://localhost

            [storage-policy:1]
            write = test1
            ''')

        conf = Configuration.from_stream(conf)

        read_set = conf.get_policy(1).read_set
        if len(read_set) != 0:
            raise self.failureException(
                '%r is not empty' % read_set)



# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Utility functions to work with `splice`'''

import errno
import fcntl
import logging
import os

import eventlet.hubs

try:
    import swift.common.splice
    HAS_NEW_SPLICE = True
except ImportError:
    import swift.common.utils
    HAS_NEW_SPLICE = False

# From `bits/fcntl-linux.h`
F_GETPIPE_SZ = 1032
MAX_PIPE_SIZE = None
MAX_PIPE_SIZE_2_6_34 = 16 * 4096


def splice_socket_to_socket(fd_in, fd_out, length=None):
    global MAX_PIPE_SIZE

    if MAX_PIPE_SIZE is None:
        try:
            with open('/proc/sys/fs/pipe-max-size', 'r') as fd:
                MAX_PIPE_SIZE = int(fd.read().strip())
        except (IOError, ValueError):
            logging.getLogger(__name__).exception(
                'Unable to read max_pipe_size')

            MAX_PIPE_SIZE = 0

    if HAS_NEW_SPLICE:
        flags = swift.common.splice.splice.SPLICE_F_MOVE | \
            swift.common.splice.splice.SPLICE_F_NONBLOCK | \
            swift.common.splice.splice.SPLICE_F_MORE
    else:
        flags = swift.common.utils.SPLICE_F_MOVE | \
            swift.common.utils.SPLICE_F_NONBLOCK | \
            swift.common.utils.SPLICE_F_MORE

    rpipe, wpipe = os.pipe()

    try:
        max_size = 0
        if MAX_PIPE_SIZE > 0:
            max_size = fcntl.fcntl(rpipe, swift.common.utils.F_SETPIPE_SZ,
                                   MAX_PIPE_SIZE)
        else:
            try:
                max_size = fcntl.fcntl(rpipe, F_GETPIPE_SZ)
            except IOError as exc:
                if exc.errno == errno.EINVAL:
                    # Most likely a pre-2.6.35 kernel (RHEL6)
                    max_size = MAX_PIPE_SIZE_2_6_34
                else:
                    raise

        assert max_size != 0, 'Calculating max_size failed'

        while (True if length is None else length > 0):
            if length is None:
                max_read_size = max_size
            else:
                max_read_size = min(max_size, length)

            if HAS_NEW_SPLICE:
                try:
                    (read, _, _) = swift.common.splice.splice(fd_in, None,
                                                              wpipe, None,
                                                              max_read_size,
                                                              flags)
                except (IOError, OSError) as exc:
                    if exc.errno == errno.EWOULDBLOCK:
                        read = None
                    else:
                        raise
            else:
                read = swift.common.utils.splice(fd_in, 0, wpipe, 0,
                                                 max_read_size, flags)

            if read is None:
                # EAGAIN
                eventlet.hubs.trampoline(fd_in, read=True)
                continue

            if read == 0:
                # EOF
                break

            todo = read

            while todo > 0:
                if HAS_NEW_SPLICE:
                    try:
                        (written, _, _) = swift.common.splice.splice(rpipe,
                                                                     None,
                                                                     fd_out,
                                                                     None,
                                                                     todo,
                                                                     flags)
                    except (IOError, OSError) as exc:
                        if exc.errno == errno.EWOULDBLOCK:
                            written = None
                        else:
                            raise
                else:
                    written = swift.common.utils.splice(rpipe, 0, fd_out, 0,
                                                        todo, flags)

                if written is None:
                    # EAGAIN
                    eventlet.hubs.trampoline(fd_out, write=True)
                    continue

                if written == 0:
                    raise ValueError('I/O operation on closed socket')

                todo -= written

            if length is not None:
                length -= read
    finally:
        os.close(rpipe)
        os.close(wpipe)



# Copyright (c) 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Command-line utilities.'''

try:
    import argparse
except ImportError:
    argparse = None
import ConfigParser
import pkg_resources
import sys

from swift_scality_backend.policy_configuration import Configuration
from swift_scality_backend.policy_configuration import ConfigurationError
from swift_scality_backend.policy_configuration import \
    DEFAULT_CONFIGURATION_PATH
from swift_scality_backend.policy_configuration import StoragePolicy


def print_error(prefix, exn):
    '''Generic exception message printer.'''

    sys.stderr.write('%s: %s\n' % (prefix, exn.message))


def storage_policy_lint(args):
    '''Implementation of the 'storage-policy-lint' subcommand.'''

    try:
        Configuration.from_stream(args.config)
        return 0
    except ConfigurationError as exn:
        print_error('Configuration error', exn)
    except ConfigParser.Error as exn:
        print_error('Parsing error', exn)
    except Exception as exn:
        print_error('Error', exn)

    return 1


def storage_policy_query(args):
    '''Implementation of the 'storage-policy-query' subcommand.'''

    try:
        config = Configuration.from_stream(args.config)
        policy = config.get_policy(args.policy_index)
    except Exception as exn:
        print_error('Error', exn)
        return 1

    print 'Query'
    print '-----'
    print 'Policy index:', args.policy_index
    print 'Action:', args.action
    print 'Location hints:', args.locations
    print

    print 'Result'
    print '------'
    result = policy.lookup(args.action, args.locations)
    for (idx, endpoints) in enumerate(result):
        print '%d: %r' % (idx, list(sorted(str(e) for e in endpoints)))

    return 0


def main(args=None):
    '''Main entry point.'''

    if not argparse:
        if sys.version_info < (2, 7):
            sys.stderr.write('This tool requires `argparse` or Python 2.7\n')
            sys.exit(2)
        else:
            raise ImportError('No module named argparse')

    parser = argparse.ArgumentParser(
        description='Scality Swift back-end utilities')
    subparsers = parser.add_subparsers(
        title='subcommands')

    try:
        package = pkg_resources.get_distribution('swift-scality-backend')
        version = '%s %s' % (package.project_name, package.version)
    except pkg_resources.DistributionNotFound:
        version = 'unknown'

    parser.add_argument('--version', action='version', version=version)

    # storage-policy-lint parser
    parser_sp_lint = subparsers.add_parser('storage-policy-lint')
    parser_sp_lint.set_defaults(func=storage_policy_lint)
    parser_sp_lint.add_argument(
        '-c', '--config',
        dest='config', action='store', type=argparse.FileType('r'),
        help='storage policy configuration to lint', metavar='FILE',
        default=DEFAULT_CONFIGURATION_PATH)

    # storage-policy-query parser
    parser_sp_query = subparsers.add_parser('storage-policy-query')
    parser_sp_query.set_defaults(func=storage_policy_query)
    parser_sp_query.add_argument(
        '-c', '--config',
        dest='config', action='store', type=argparse.FileType('r'),
        help='storage policy configuration to query', metavar='FILE',
        default=DEFAULT_CONFIGURATION_PATH)
    parser_sp_query.add_argument(
        'action',
        choices=[StoragePolicy.READ, StoragePolicy.WRITE],
        help='action to emulate', metavar='ACTION')
    parser_sp_query.add_argument(
        'policy_index',
        action='store', type=int,
        help='storage policy index', metavar='INDEX')
    parser_sp_query.add_argument(
        'locations',
        help='location hints', metavar='LOCATION',
        nargs='*')

    # Go
    args2 = parser.parse_args(args=args)
    rc = args2.func(args2)

    sys.exit(rc)


if __name__ == '__main__':
    main()



# Copyright (c) 2014, 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of functions that helps running unit tests"""

import functools
import re
import unittest

import eventlet
import mock
import nose.plugins.skip

from swift_scality_backend.http_utils import ClientCollection

from scality_sproxyd_client.sproxyd_client import SproxydClient


def skipIf(condition, reason):
    """
    A `skipIf` decorator.

    Similar to `unittest.skipIf`, for Python 2.6 compatibility.
    """
    def decorator(test_item):
        @functools.wraps(test_item)
        def wrapped(*args, **kwargs):
            if condition:
                raise nose.plugins.skip.SkipTest(reason)
            else:
                return test_item(*args, **kwargs)
        return wrapped
    return decorator


def assertRaisesRegexp(expected_exception, expected_regexp,
                       callable_obj, *args, **kwargs):
    """Asserts that the message in a raised exception matches a regexp."""
    try:
        callable_obj(*args, **kwargs)
    except expected_exception as exc_value:
        if not re.search(expected_regexp, str(exc_value)):
            # We accept both `string` and compiled regex object as 2nd
            # argument to assertRaisesRegexp
            pattern = getattr(expected_regexp, 'pattern', expected_regexp)
            raise unittest.TestCase.failureException(
                '"%s" does not match "%s"' %
                (pattern, str(exc_value)))
    else:
        if hasattr(expected_exception, '__name__'):
            excName = expected_exception.__name__
        else:
            excName = str(expected_exception)
        raise unittest.TestCase.failureException("%s not raised" % excName)


def assertRegexpMatches(text, expected_regexp, msg=None):
    '''Asserts the text matches the regular expression.'''

    if isinstance(expected_regexp, basestring):
        expected_regexp = re.compile(expected_regexp)

    if not expected_regexp.search(text):
        msg = msg or "Regexp didn't match"
        msg = '%s: %r not found in %r' % (msg, expected_regexp.pattern, text)
        raise unittest.TestCase.failureException(msg)


def make_client_collection(endpoints=None, conn_timeout=None,
                           read_timeout=None, logger=None):
    '''Construct an `SproxydClient` instance using default values.'''

    def maybe(default, value):
        return value if value is not None else default

    endpoints = maybe(['http://localhost:81/proxy/chord/'], endpoints)
    conn_timeout = maybe(10.0, conn_timeout)
    read_timeout = maybe(3.0, read_timeout)
    logger = maybe(mock.Mock(), logger)

    client = SproxydClient(
        endpoints=endpoints, conn_timeout=conn_timeout,
        read_timeout=read_timeout, logger=logger)

    return ClientCollection(read_clients=[client], write_clients=[client])


class WSGIServer(object):
    """Start a WSGI Web server."""
    def __init__(self, application):
        self.application = application

    def __enter__(self):
        self._server = eventlet.listen(('127.0.0.1', 0))
        (self.ip, self.port) = self._server.getsockname()
        self._thread = eventlet.spawn(eventlet.wsgi.server, self._server,
                                      self.application)
        return self

    def __exit__(self, exc_ty, exc_val, tb):
        try:
            self._thread.kill()
        finally:
            self._server.close()



#!/usr/bin/env python
# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import distutils.spawn
import setuptools
import subprocess
import sys

import swift_scality_backend


def get_version():
    def has_git():
        return distutils.spawn.find_executable('git') is not None

    def is_git_clone():
        cmd = ['git', 'rev-parse', '--show-toplevel']

        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        return proc.wait() == 0

    if hasattr(subprocess, 'check_output'):
        check_output = subprocess.check_output
    else:
        def check_output(*args, **kwargs):
            proc = subprocess.Popen(stdout=subprocess.PIPE, *args, **kwargs)
            output, _ = proc.communicate()
            rc = proc.poll()
            if rc != 0:
                raise subprocess.CalledProcessError(
                    rc, kwargs.get('args', args[0]), output=output)
            return output

    def get_git_version():
        prefix = 'swift-scality-backend-'
        cmd = ['git', 'describe', '--tags', '--dirty', '--always',
               '--match', '%s*' % prefix]

        result = check_output(cmd).strip()
        if sys.version_info >= (3, 0):
            result = str(result, 'utf-8')
        assert result.startswith(prefix)

        return result[len(prefix):]

    if has_git() and is_git_clone():
        return get_git_version()
    else:
        return '999'


setuptools.setup(
    name='swift-scality-backend',
    version='0.4.4',
    description='Scality Ring backend for OpenStack Swift',
    url='http://www.scality.com/',
    author='Scality Openstack Engineering Team',
    author_email='openstack-eng@scality.com',
    license='Apache License (2.0)',
    packages=['swift_scality_backend'],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7'],
    install_requires=swift_scality_backend.__requires__,
    entry_points={
        'paste.app_factory': [
            'sproxyd_object=swift_scality_backend.server:app_factory'],
        'console_scripts': [
            'swift-scality-backend=swift_scality_backend.cli:main'],
    },
)



import os.path

import fabric.contrib.files

from fabric.api import cd, sudo


def render(directory, filenames, local_path_prefix, content):
    """
    Render a list of templates in a certain directory and upload them.

    :param directory: local directory holding templates
    :type directory: string
    :param filenames: files to render from local directory
    :type filenames: list of strings
    :param local_path_prefix: prefix to strip from the directory path to
        obtain remote path
    :type local_path_prefix: string
    :param content: content for string interpolation on templates
    :type content: dict
    """
    remote_directory = directory[len(local_path_prefix):]
    sudo('mkdir -p {0:s}'.format(remote_directory))
    for filename in filenames:
        local_path = os.path.join(directory, filename)
        fabric.contrib.files.upload_template(
            filename=local_path,
            destination=os.path.join(remote_directory, filename),
            context=content,
            use_sudo=True,
        )


def build_object_ring(swift_user, name, devices, part_power=10, replicas=3):
    """
    Build a swift object ring by invocation of `swift-ring-builder`.

    :param swift_user: the user running swift
    :type swift_user: string
    :param name: name of the ring file to create, eg. `account.builder`
    :type name: string
    :param devices: list of fully qualified device names
        z<zone>-<ip>:<port>/<device>_<meta>
    :type devices: list of strings
    :param part_power: 2^<part_power> of partitions in the ring (optional)
    :type part_power: int
    :param replicas: replication factor
    :type replicas: int
    """
    # swift-ring-builder <file> create <part_power> <replicas> <min_part_hours>
    create_cmd = "swift-ring-builder {0:s} create {1:d} {2:d} 1"
    # swift-ring-builder <file> add z<zone>-<ip>:<port>/<dev>_<meta> <weight>
    add_cmd = "swift-ring-builder {0:s} add {1:s} 1"
    with cd('/etc/swift'):
        sudo(create_cmd.format(name, part_power, replicas), user=swift_user)

        for device in devices:
            sudo(add_cmd.format(name, device), user=swift_user)

        sudo('swift-ring-builder {0:s} rebalance'.format(name),
             user=swift_user)


def apt_get(packages):
    """
    Install packages through apt-get.

    :param packages: list of packages for installation
    :type packages: list of strings
    """
    sudo('apt-get update')
    sudo('apt-get install -y {0:s}'.format(' '.join(packages)))



# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift_scality_backend.splice_utils`'''

import errno
import fcntl
import os
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

import eventlet
import mock

import swift.common.utils

import swift_scality_backend.splice_utils

import utils


def _test_splice_socket_to_socket(test_length):
    # Linux 2.6 (RHEL6) doesn't have this procfs entry
    # Whilst the code in `splice_utils` handles this gracefully, the tests used
    # to fail because of this.
    try:
        with open('/proc/sys/fs/pipe-max-size', 'r') as fd:
            max_size = int(fd.read().strip())
    except IOError as exc:
        if exc.errno == errno.ENOENT:
            (rpipe, wpipe) = os.pipe()
            try:
                max_size = fcntl.fcntl(
                    rpipe, swift_scality_backend.splice_utils.F_GETPIPE_SZ)
            except IOError as exc:
                if exc.errno == errno.EINVAL:
                    max_size = \
                        swift_scality_backend.splice_utils.MAX_PIPE_SIZE_2_6_34
                else:
                    raise

            finally:
                os.close(rpipe)
                os.close(wpipe)
        else:
            raise

    orig_message = 'Hello, world!' * max_size

    if test_length:
        test_message = orig_message[:-2]
    else:
        test_message = orig_message

    test_message_length = len(test_message)

    server1 = eventlet.listen(('127.0.0.1', 0))
    addr1 = server1.getsockname()

    server2 = eventlet.listen(('127.0.0.1', 0))
    addr2 = server2.getsockname()

    result = StringIO.StringIO()

    def run_server1(sock):
        (client, addr) = sock.accept()

        remote = eventlet.connect(addr2)

        if test_length:
            swift_scality_backend.splice_utils.splice_socket_to_socket(
                client.fileno(), remote.fileno(), length=test_message_length)
        else:
            swift_scality_backend.splice_utils.splice_socket_to_socket(
                client.fileno(), remote.fileno())

    def run_server2(sock):
        (client, addr) = sock.accept()

        while True:
            data = client.recv(max_size)

            if len(data) == 0:
                break

            result.write(data)

    thread1 = eventlet.spawn(run_server1, server1)
    thread2 = eventlet.spawn(run_server2, server2)

    client = eventlet.connect(addr1)
    client.sendall(orig_message)
    client.close()

    thread1.wait()
    thread2.wait()

    assert result.getvalue() == test_message


try:
    import swift.common.splice
    HAS_SPLICE = swift.common.splice.splice.available
except ImportError:
    try:
        HAS_SPLICE = swift.common.utils.system_has_splice
    except AttributeError:
        HAS_SPLICE = False


@utils.skipIf(not HAS_SPLICE, "No `splice` support")
def test_splice_socket_to_socket():
    return _test_splice_socket_to_socket(test_length=False)


@utils.skipIf(not HAS_SPLICE, "No `splice` support")
def test_splice_socket_to_socket_bounded():
    return _test_splice_socket_to_socket(test_length=True)


@utils.skipIf(not HAS_SPLICE, "No `splice` support")
def test_splice_no_pipe_max_size():
    '''Test absence of `/proc/sys/fs/pipe-max-size`.'''

    open_mock = mock.mock_open()

    default_open = open

    def fake_open(name, *args, **kwargs):
        if name == '/proc/sys/fs/pipe-max-size':
            raise IOError(errno.ENOENT, 'No such file or directory')
        else:
            return default_open(name, *args, **kwargs)

    open_mock.side_effect = fake_open

    with mock.patch('__builtin__.open', open_mock):
        with mock.patch('fcntl.fcntl', side_effect=fcntl.fcntl) as mock_fcntl:
            swift_scality_backend.splice_utils.MAX_PIPE_SIZE = None

            try:
                _test_splice_socket_to_socket(test_length=True)

                mps = swift_scality_backend.splice_utils.MAX_PIPE_SIZE
            finally:
                swift_scality_backend.splice_utils.MAX_PIPE_SIZE = None

            assert mps == 0, 'Unexpected MAX_PIPE_SIZE value: %d' % mps

            assert mock_fcntl.call_count == 2, 'Unexpected fcntl call count'

            for (args, kwargs) in mock_fcntl.call_args_list:
                cmd = args[1]
                assert \
                    cmd == swift_scality_backend.splice_utils.F_GETPIPE_SZ, \
                    'Unexpected fcntl call: %d' % cmd
